from gunfolds.utils import ecj
from gunfolds.utils import bfutils as bfu
from gunfolds.conversions import graph2adj, adjs2graph
from gunfolds.utils import graphkit as gk
import numpy as np
from progressbar import ProgressBar, Percentage
from scipy import linalg, optimize
from statsmodels.tsa.api import VAR
from sympy.matrices import SparseMatrix


def symchol(M):  # symbolic Cholesky
    """
    :param M:
    :type M:
    
    :returns: 
    :rtype: 
    """
    B = SparseMatrix(M)
    t = B.row_structure_symbolic_cholesky()
    B = np.asarray(B)*0
    for i in range(B.shape[0]):
        B[i, t[i]] = 1
    return B


def G2SVAR(G):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graph)
    
    :returns: 
    :rtype: 
    """
    n = len(G)
    A, B = npG2SVAR(G)
    P, L, U = linalg.lu(B)
    A = linalg.inv(L).tolist()
    B = B.tolist()
    A = listplace(A, 0.0, 0.0)
    for i in range(0, n):
        A[i][i] = 1
    B = listplace(B, 0.0, 'e')
    for i in range(0, n):
        B[i][i] = 'e'
    return A, B, P


def G2AH(G):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graph)
    
    :returns: 
    :rtype: 
    """
    n = len(G)
    A, B = npG2SVAR(G)
    P, L, U = linalg.lu(B)
    A = linalg.inv(L).tolist()
    B = B.tolist()
    A = listplace(A, 0.0, 0.0)
    for i in range(0, n):
        A[i][i] = 1
    B = listplace(B, 0.0, 'e')
    for i in range(0, n):
        B[i][i] = 'e'
    return A, B, P


def bnf2CG(fname):
    """
    :param fname:
    :type fname:
    
    :returns: 
    :rtype: 
    """
    d = eval(open(fname).read())
    G = {}
    for v in d:
        G[v] = {u: 1 for u in d[v]['pars']}
    G = ecj.tr(G)
    for v in G:
        ld = {u: 1 for u in G[v]}
        G[v] = ld
    return G


def npG2SVAR(G):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graph)
    
    :returns: 
    :rtype: 
    """
    n = len(G)
    A = [[0]*n]*n
    B = [[0]*n]*n
    for i in range(n):
        B[i][i] = 1

    for v in G:
        for w in G[v]:
            if G[v][w] in (1, 3):
                A[w-1][v-1] = 1
            if G[v][w] in (2, 3):
                B[w-1][v-1] = 1
    A = np.asarray(A)
    B = symchol(B)
    return A, B


def x2M(x, A, B, aidx, bidx):
    """
    :param x:
    :type x:
    
    :param A:
    :type A:
    
    :param B:
    :type B:
    
    :param aidx:
    :type aidx:
    
    :param bidx:
    :type bidx:
    
    :returns: 
    :rtype: 
    """
    A[aidx] = x[:len(aidx[0])]
    B[bidx] = x[len(aidx[0]):]
    # B[(bidx[1],bidx[0])] = x[len(aidx[0]):]
    return A, B


def nllf(x, A, B, Y, aidx, bidx):  # negative log likelihood
    """
    :param x:
    :type x:
    
    :param A:
    :type A:
    
    :param B:
    :type B:
    
    :param Y:
    :type Y:
    
    :param aidx:
    :type aidx:
    
    :param bidx:
    :type bidx:
    
    :returns: 
    :rtype: 
    """
    A, B = x2M(x, A, B, aidx, bidx)
    T = Y.shape[1]
    X = Y[:, 1:] - np.dot(A, Y[:, :-1])
    ldB = T*np.log(abs(1./linalg.det(B)))
    return ldB + 0.5*np.trace(np.dot(np.dot(B.T, B), np.dot(X, X.T)))


def nllf2(x, A, B, YY, XX, YX, T, aidx, bidx):  # negative log likelihood
    """
    :param x:
    :type x:
    
    :param A:
    :type A:
    
    :param B:
    :type B:
    
    :param YY:
    :type YY:
    
    :param XX:
    :type XX:
    
    :param YX:
    :type YX:
    
    :param T:
    :type T:
    
    :param aidx:
    :type aidx:
    
    :param bidx:
    :type bidx:
    
    :returns: 
    :rtype: 
    """
    A, B = x2M(x, A, B, aidx, bidx)
    AYX = np.dot(A, YX.T)
    S = YY - AYX - AYX.T + np.dot(np.dot(A, XX), A.T)
    ldB = T*np.log(abs(1./linalg.det(B)))
    return 0.5*np.dot(np.dot(B.T, B).T.flat, S.flat) + ldB
    # return ldB + 0.5*np.trace( np.dot(np.dot(B.T, B), S))


def VARbic(nllf, K, T):
    """
    :param nllf:
    :type nllf:
    
    :param K:
    :type K:
    
    :param T:
    :type T:
    
    :returns: 
    :rtype: 
    """
    return 2*nllf + K*np.log(T)


def listplace(l, a, b):
    """
    :param l:
    :type l:
    
    :param a:
    :type a:
    
    :param b:
    :type b:
    
    :returns: 
    :rtype: 
    """
    return [listplace(x, a, b) if not np.isscalar(x) else b if x != a else x for x in l]

# -------------------------------------------------------------------
# data generation
# -------------------------------------------------------------------


def randweights(n, c=0.1, factor=9):
    """
    :param n:
    :type n:
    
    :param c:
    :type c: float
    
    :param factor:
    :type factor: (guess)integer
    
    :returns: 
    :rtype: 
    """
    rw = np.random.randn(n)
    idx = np.where(abs(rw) < factor*c)
    if idx:
        rw[idx] = rw[idx]+np.sign(rw[idx])*c*factor
    return rw


def transitionMatrix(cg, minstrength=0.1):
    """
    :param cg:
    :type cg:
    
    :param minstrength:
    :type minstrength: float
    
    :returns: 
    :rtype: 
    """
    A = graph2adj(cg)
    edges = np.where(A == 1)
    A[edges] = randweights(edges[0].shape[0], c=minstrength)
    l = linalg.eig(A)[0]
    c = 0
    pbar = ProgressBar(widgets=['Searching for weights: ', Percentage(), ' '], maxval=10000).start()
    while max(l*np.conj(l)) > 1:
        A[edges] = randweights(edges[0].shape[0], c=c)
        c += 1
        l = linalg.eig(A)[0]
        pbar.update(c)
    pbar.finish()
    return A


def sampleWeights(n, minstrength=0.1):
    """
    :param n:
    :type n:
    
    :param minstrength:
    :type minstrength: float
    
    :returns: 
    :rtype: 
    """
    r = np.randn(n)
    s = minstrength/np.min(np.abs(r))
    r = s*r
    return r


def transitionMatrix2(cg, minstrength=0.1):
    """
    :param cg:
    :type cg:
    
    :param minstrength:
    :type minstrength: float
    
    :returns: 
    :rtype: 
    """
    A = graph2adj(cg)
    edges = np.where(A == 1)
    A[edges] = sampleWeights(edges[0].shape[0], minstrength=minstrength)
    l = linalg.eig(A)[0]
    c = 0
    pbar = ProgressBar(widgets=['Searching for weights: ', Percentage(), ' '], maxval=10000).start()
    while max(l*np.conj(l)) > 1:
        A[edges] = sampleWeights(edges[0].shape[0], minstrength=minstrength)
        c += 1
        l = linalg.eig(A)[0]
        if c > pbar.maxval:
            raise ValueError
        pbar.update(c)
    pbar.finish()
    return A


def transitionMatrix3(cg, x0=None, minstrength=0.1):
    """
    :param cg:
    :type cg:
    
    :param x0:
    :type x0:
    
    :param minstrength:
    :type minstrength: float
    
    :returns: 
    :rtype: 
    """
    A = graph2adj(cg)
    edges = np.where(A == 1)

    try:
        x = x0
    except AttributeError:
        A = initRandomMatrix(A, edges)
        x = A[edges]

    def objective(x):
        A[edges] = np.real(x)
        l = linalg.eig(A)[0]
        m = np.max(np.real(l*np.conj(l)))-0.99
        n = np.min(np.min(np.abs(x)), minstrength)-minstrength
        return m*m + 0.1*n*n

    o = np.zeros(len(edges))
    while np.min(np.abs(o[0])) < 0.8*minstrength:
        rpt = True
        while rpt:
            try:
                try:
                    o = optimize.fmin_bfgs(objective, x,
                                           gtol=1e-10, maxiter=100,
                                           disp=False, full_output=True)
                    A[edges] = np.real(o[0])
                    l = linalg.eig(A)[0]
                    if np.max(np.real(l*np.conj(l))) < 1:
                        rpt = False

                except:
                    rpt = True
            except Warning:
                x = np.randn(len(edges[0]))
                rpt = True
    A[edges] = np.real(o[0])
    return A


def initRandomMatrix(A, edges, distribution='beta'):
    """
    possible distributions:
    flat
    flatsigned
    beta
    normal
    uniform

    :param A:
    :type A:
    
    :param edges:
    :type edges:
    
    :param distribution: (GUESS)distribution from which to sample the weights. Available 
     options are flat, flatsigned, beta, normal, uniform
    :type distribution: string

    :returns: 
    :rtype: 
    """
    def init():
        if distribution == 'flat':
            x = np.ones(len(edges[0]))
        elif distribution == 'flatsigned':
            x = np.sign(np.randn(len(edges[0])))*np.ones(len(edges[0]))
        elif distribution == 'beta':
            x = np.random.beta(0.5, 0.5, len(edges[0]))*3-1.5
        elif distribution == 'normal':
            x = np.randn(len(edges[0]))
        elif distribution == 'uniform':
            x = np.sign(np.randn(len(edges[0])))*np.rand(len(edges[0]))
        else:
            raise ValueError('Wrong option!')
        return x

    def eigenvalue(A):
        l = linalg.eig(A)[0]
        s = np.max(np.real(l*np.conj(l)))
        return s

    x = init()
    A[edges] = x
    s = eigenvalue(A)
    alpha = np.random.rand()*(0.99-0.8)+0.8
    A = A/(alpha*s)
    return A


def transitionMatrix4(g, minstrength=0.1, distribution='normal', maxtries=1000):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graph)
    
    :param minstrength:
    :type minstrength: float
    
    :param distribution: (GUESS)distribution from which to sample the weights. Available 
     options are flat, flatsigned, beta, normal, uniform
    :type distribution: string
    
    :param maxtries:
    :type maxtries: (guess)integer
    
    :returns: 
    :rtype: 
    """
    A = graph2adj(g)
    edges = np.where(A == 1)
    s = 2.0
    c = 0
    pbar = ProgressBar(widgets=['Searching for weights: ',
                                Percentage(), ' '],
                       maxval=maxtries).start()
    while s > 1.0:
        minstrength -= 0.001
        A = initRandomMatrix(A, edges, distribution=distribution)
        x = A[edges]
        delta = minstrength/np.min(np.abs(x))
        A[edges] = delta*x
        l = linalg.eig(A)[0]
        s = np.max(np.real(l*np.conj(l)))
        c += 1
        if c > maxtries:
            return None
        pbar.update(c)
    pbar.finish()

    return A


def drawsamplesLG(A, nstd=0.1, samples=100):
    """
    :param A:
    :type A:
    
    :param nstd:
    :type nstd: float
    
    :param samples:
    :type samples: integer
    
    :returns: 
    :rtype: 
    """
    n = A.shape[0]
    data = np.zeros([n, samples])
    data[:, 0] = nstd*np.random.randn(A.shape[0])
    for i in range(1, samples):
        data[:, i] = A @ data[:, i-1] + nstd*np.random.randn(A.shape[0])
    return data


def drawsamplesMA(A, nstd=0.1, samples=100, order=5):
    """
    :param A:
    :type A:
    
    :param nstd:
    :type nstd: float
    
    :param samples:
    :type samples: integer
    
    :param order:
    :type order: integer
    
    :returns: 
    :rtype: 
    """
    n = A.shape[0]
    data = np.zeros([n, samples])
    data[:, 0] = nstd*np.random.randn(A.shape[0])
    for i in range(1, samples):
        if i > order:
            result = 0
            for j in range(order):
                result += np.dot(1/(j+1)*A, data[:, i-1-j]) \
                  + nstd*np.dot(1/(j+1)*A, np.random.randn(A.shape[0]))
            data[:, i] = result
        else:
            data[:, i] = A @ data[:, i-1] \
              + nstd*np.random.randn(A.shape[0])
    return data


def getAgraph(n, mp=2, st=0.5, verbose=True):
    """
    :param n:
    :type n:
    
    :param mp:
    :type mp: (guess)integer
    
    :param st:
    :type st: float
    
    :param verbose:
    :type verbose: boolean
    
    :returns: 
    :rtype: 
    """
    keeptrying = True
    while keeptrying:
        G = gk.rnd_CG(n, maxindegree=mp, force_connected=True)
        try:
            A = transitionMatrix2(G, minstrength=st)
            keeptrying = False
        except ValueError as e:
            if verbose:
                print("!!! Unable to find strong links for a stable matrix !!!")
                print("*** trying a different graph")
    return {'graph':      G,
            'transition': A,
            'converges':  len(bfu.call_undersamples(G))}


def getAring(n, density=0.1, st=0.5, verbose=True, dist='flatsigned'):
    """
    :param n:
    :type n:
    
    :param density: (guess)ratio of total nodes to n^2 possible nodes
    :type density: float
    
    :param st:
    :type st: float
    
    :param verbose:
    :type verbose: boolean
    
    :param dist:
    :type dist: string
    
    :returns: 
    :rtype: 
    """
    keeptrying = True
    plusedges = bfu.dens2edgenum(density, n)
    while keeptrying:
        G = gk.ringmore(n, plusedges)
        try:
            A = transitionMatrix4(G, minstrength=st, distribution=dist)
            try:
                s = A.shape
                keeptrying = False
            except AttributeError:
                keeptrying = True
        except ValueError:
            if verbose:
                print("!!! Unable to find strong links for a stable matrix !!!")
                print("*** trying a different graph")
    return {'graph':      G,
            'transition': A,
            'converges':  len(bfu.call_undersamples(G))}


# -------------------------------------------------------------------
# estimation
# -------------------------------------------------------------------

def scoreAGraph(G, data, x0=None):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graph)
    
    :param data:
    :type data:
    
    :param x0:
    :type x0:
    
    :returns: 
    :rtype: 
    """
    A, B = npG2SVAR(G)
    K = np.sum(abs(A)+abs(B))
    a_idx = np.where(A != 0)
    b_idx = np.where(B != 0)
    if x0:
        o = optimize.fmin_bfgs(nllf, x0, args=(A, B, data, a_idx, b_idx),
                               disp=False, full_output=True)
    else:
        o = optimize.fmin_bfgs(nllf, np.randn(K),
                               args=(np.double(A), np.double(B),
                                     data, a_idx, b_idx),
                               disp=False, full_output=True)
    VARbic(o[1], K, data.shape[1])


def estimateG(G, YY, XX, YX, T, x0=None):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graph)
    
    :param YY:
    :type YY:
    
    :param XX:
    :type XX:
    
    :param YX:
    :type YX:
    
    :param T:
    :type T:
    
    :param x0:
    :type x0:
    
    :returns: 
    :rtype: 
    """
    A, B = npG2SVAR(G)
    K = np.sum(abs(A)+abs(B))
    a_idx = np.where(A != 0)
    b_idx = np.where(B != 0)
    try:
        s = x0.shape
        x = x0
    except AttributeError:
        x = np.randn(K)
    o = optimize.fmin_bfgs(nllf2, x,
                           args=(np.double(A), np.double(B),
                                 YY, XX, YX, T, a_idx, b_idx),
                           disp=False, full_output=True)
    A, B = x2M(o[0], np.double(A), np.double(B), a_idx, b_idx)
    return A, B


def data2AB(data, x0=None):
    """
    :param data:
    :type data:
    
    :param x0:
    :type x0:
    
    :returns: 
    :rtype: 
    """
    n = data.shape[0]
    T = data.shape[1]
    YY = np.dot(data[:, 1:], data[:, 1:].T)
    XX = np.dot(data[:, :-1], data[:, :-1].T)
    YX = np.dot(data[:, 1:], data[:, :-1].T)

    model = VAR(data.T)
    r = model.fit(1)
    A = r.coefs[0, :, :]

    # A = np.ones((n,n))
    B = np.ones((n, n))
    np.fill_diagonal(B, 0)
    B[np.triu_indices(n)] = 0
    K = np.sum(abs(B)).astype(int)  # abs(A)+abs(B)))

    a_idx = np.where(A != 0)
    b_idx = np.where(B != 0)
    np.fill_diagonal(B, 1)

    try:
        s = x0.shape
        x = x0
    except AttributeError:
        x = np.r_[A.flatten(), 0.1*np.random.randn(K)]
    o = optimize.fmin_bfgs(nllf2, x,
                           args=(np.double(A), np.double(B),
                                 YY, XX, YX, T, a_idx, b_idx),
                           gtol=1e-12, maxiter=500,
                           disp=False, full_output=True)
    A, B = x2M(o[0], np.double(A), np.double(B), a_idx, b_idx)
    B = B+B.T
    return A, B


def amap(f, a):
    """
    :param f:
    :type f:
    
    :param a:
    :type a:
    
    :returns: 
    :rtype: 
    """
    v = np.vectorize(f)
    return v(a)


def AB2intAB(A, B, th=0.09):
    """
    :param A:
    :type A:
    
    :param B:
    :type B:
    
    :param th: (GUESS)threshold for discarding edges in A and B
    :type th: float
    
    :returns: 
    :rtype: 
    """
    A[amap(lambda x: abs(x) > th, A)] = 1
    A[amap(lambda x: abs(x) < 1, A)] = 0
    B[amap(lambda x: abs(x) > th, B)] = 1
    B[amap(lambda x: np.abs(x) < 1, B)] = 0
    np.fill_diagonal(B, 0)
    return A, B


def data2graph(data, x0=None, th=0.0):
    """
    :param data:
    :type data:
    
    :param x0:
    :type x0:
    
    :returns: 
    :rtype: 
    """
    A, B = data2AB(data, x0=x0)
    Ab = A.copy()
    Bb = B.copy()
    A, B = AB2intAB(A, B, th=th)
    return adjs2graph(A, B), Ab, Bb


def data2VARgraph(data, pval=0.05):
    """
    :param data:
    :type data:
    
    :param pval:
    :type pval: float
    
    :returns: 
    :rtype: 
    """
    model = VAR(data.T)
    r = model.fit(1)
    A = r.coefs[0, :, :]
    n = A.shape[0]
    g = {i: {} for i in range(1, n+1)}

    for i in range(n):
        for j in range(n):
            if np.abs(A[j, i]) > pval:
                g[i+1][j+1] = 1
    return g


# this is for the SAT solver project
def stableVAR(n, density=0.1, dist='beta'):
    """
    This function keeps trying to create a random graph and a random
    corresponding transition matrix until it succeeds.

    :param n: number of nodes in the graph
    :type n: (guess)integer
    
    :param density: ratio of total nodes to n^2 possible nodes
    :type density: (guess)float
        
    :param dist: distribution from which to sample the weights. Available 
     options are flat, flatsigned, beta, normal, uniform
    :type dist: (guess)string
    
    :returns: 
    :rtype: 
    """
    np.random.seed()
    sst = 0.9
    r = None
    while not r:
        r = getAring(n, density, sst, False, dist=dist)
        if sst < 0.03:
            sst -= 0.001
        else:
            sst -= 0.01
        if sst < 0:
            sst = 0.02
    return r['graph'], r['transition']


def genData(n, rate=2, density=0.1, burnin=100, ssize=2000, noise=0.1, dist='beta'):
    """
    Given a number of nodes this function randomly generates a ring
    SCC and the corresponding stable transition matrix. It tries until
    succeeds and for some graph densities and parameters of the
    distribution of transition matrix values it may take
    forever. Please play with the dist parameter to stableVAR. Then
    using this transition matrix it generates `ssize` samples of data
    and undersamples them by `rate` discarding the `burnin` number of
    samples at the beginning.

    :param n: number of nodes in the desired graph
    :type n: (guess)integer
    
    :param rate: undersampling rate (1 - no undersampling)
    :type rate: integer
    
    :param density: density of the graph to be generted
    :type density: (guess) float
    
    :param burnin: number of samples to discard since the beginning of VAR sampling
    :type burnin: integer
    
    :param ssize: how many samples to keep at the causal sampling rate
    :type ssize: (guess)integer
    
    :param noise: noise standard deviation for the VAR model
    :type noise: (guess)float
    
    :param dist: (GUESS)distribution from which to sample the weights. Available 
     options are flat, flatsigned, beta, normal, uniform
    :type dist: (guess)string
    
    :returns: 
    :rtype: 
    """
    g, Agt = stableVAR(n, density=density, dist=dist)
    data = drawsamplesMA(Agt, samples=burnin + ssize * 2, nstd=noise)
    data = data[:, burnin:]
    return g, Agt, data[:, ::rate]


def estimateSVAR(data, th=0.09):
    """
    :param data:
    :type data:
    
    :param th: (GUESS)threshold for discarding edges in A and B
    :type th: (guess)float
    
    :returns: 
    :rtype: 
    """
    A, B = data2AB(data)
    A, B = AB2intAB(A, B, th=th)
    return A, B


# option #1
def randomSVAR(n, rate=2, density=0.1, th=0.09, burnin=100,
               ssize=2000, noise=0.1, dist='beta'):
    """
    Given a number of nodes this function randomly generates a ring
    SCC and the corresponding stable transition matrix. It tries until
    succeeds and for some graph densities and parameters of the
    distribution of transition matrix values it may take
    forever. Please play with the dist parameter to stableVAR. Then
    using this transition matrix it generates `ssize` samples of data
    and undersamples them by `rate` discarding the `burnin` number of
    samples at the beginning. For these data the funcion solves the
    SVAR estimation maximizing log likelihood and returns the A and B
    matrices.

    :param n: number of nodes in the desired graph
    :type n: (guess)integer
    
    :param rate: undersampling rate (1 - no undersampling)
    :type rate: integer
    
    :param density: density of the graph to be generted
    :type density: (guess)float
        
    :param th: threshold for discarding edges in A and B
    :type th: (guess)float
    
    :param burnin: number of samples to discard since the beginning of VAR sampling
    :type burnin: (guess)integer
    
    :param ssize: how many samples to keep at the causal sampling rate
    :type ssize: (guess)integer
    
    :param noise: noise standard deviation for the VAR model
    :type noise: (guess)float
    
    :param dist: (GUESS)distribution from which to sample the weights. Available 
     options are flat, flatsigned, beta, normal, uniform
    :type dist: (guess)string
    
    :returns: 
    :rtype: 
    """
    g, Agt, data = genData(n, rate=rate, density=density,
                           burnin=burnin, ssize=ssize, noise=noise, dist=dist)
    A, B = estimateSVAR(data, th=th)
    return {'graph': g,
            'rate': rate,
            'graph@rate': bfu.undersample(g, rate-1),
            'transition': Agt,
            'estimate': adjs2graph(A, B),
            'directed': A,
            'bidirected': B
            }


# option #2
def noiseData(data, noise=0.1):
    """
    :param data:
    :type data:
    
    :param noise: (GUESS)noise standard deviation for the VAR model
    :type noise: (guess)float
    
    :returns: 
    :rtype: 
    """
    h, w = data.shape
    return data + np.random.randn(h, w)*noise


def decide_absences(As):
    """
    Given a list of binary matrices returns a binary mask for absence
    and presence of edges

    :param As: a list of binary matrices
    :type As: 
    
    :returns: 
    :rtype: 
    """
    M = np.zeros(As[0].shape).astype('int')
    M[np.where(np.sum(As, axis=0) > len(As)/2.0)] = 1
    return M


def presence_probs(As):
    """
    Given a list of binary matrices returns a frequency of edge
    presence

    :param As: a list of binary matrices
    :type As:
    
    :returns: 
    :rtype: 
    """
    n = len(As)
    M = np.sum([np.zeros(As[0].shape), np.ones(As[0].shape)]+As, axis=0)
    return M/(n+2.0)


def weight_and_mask(As):
    """
    Given a list o fbinary matrices returns a weight matrix for
    presences and absences and a mask to identify which are which

    :param As: a list of binary matrices
    :type As:
    
    :returns: 
    :rtype: 
    """
    M = decide_absences(As)
    W = presence_probs(As)
    A = np.ones(M.shape) - W  # ansence probs
    A[np.where(M == 1)] = W[np.where(M == 1)]
    return (1000 * (np.log(A) - np.log(1-A))).astype('int'), M


def randomSVARs(n, repeats=100, rate=2, density=0.1, th=0.09,
                burnin=100, ssize=2000, noise=0.1, strap_noise=0.1):
    """
    does what requested - help is on the way

    :param n: number of nodes in the desired graph
    :type n: integer
    
    :param repeats: how many times to add noise and re-estiamte
    :type repeats: integer
    
    :param rate: undersampling rate (1 - no undersampling)
    :type rate: integer
    
    :param density: density of the graph to be generted
    :type density: (guess)float
    
    :param th: threshold for discarding edges in A and B
    :type th: (guess)float
    
    :param burnin: number of samples to discard since the beginning of
          VAR sampling
    :type burnin: integer
    
    :param ssize: how many samples to keep at the causal sampling rate
    :type ssize: integer
    
    :param noise: noise standard deviation for the VAR model
    :type noise: float
    
    :param strap_noise: amount of noise for bootstrapping
    :type strap_noise: float
    
    :returns: 
    :rtype: 
    """
    g, Agt, data = genData(n, rate=rate, density=density,
                           burnin=burnin, ssize=ssize, noise=noise)

    As = []
    Bs = []
    A, B = estimateSVAR(data, th=th)
    As.append(A)
    Bs.append(B)

    for i in range(repeats-1):
        A, B = estimateSVAR(noiseData(data, noise=strap_noise), th=th)
        As.append(A)
        Bs.append(B)

    A = weight_and_mask(As)
    B = weight_and_mask(Bs)

    return {'graph': g,
            'rate': rate,
            'graph@rate': bfu.undersample(g, rate-1),
            'transition': Agt,
            'directed': A,
            'bidirected': B
            }

# option #3



from gunfolds.conversions import num2CG, graph2nx
from gunfolds.utils.bfutils import undersample
from gunfolds.utils import ecj
import networkx as nx
import numpy as np
import operator
from functools import reduce

np.random.RandomState()


def has_self_loops(G):
    """
    Checks if a graph has self loop and vise versa

    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :returns: True, if the graph has self loop and vise versa
    :rtype: boolean
    """
    for u in G:
        if u in G[u]:
            return True
    return False


def randSCC(n):
    """
    Returns a random ``gunfolds`` graph with gcd>1

    :param n: number of nodes
    :type n: integer
    
    :returns: random ``gunfolds`` graph with gcd>1
    :rtype: dictionary (``gunfolds`` graphs)
    """
    G = num2CG(np.random.randint(2 ** (n ** 2)), n)
    while (len(ecj.scc(G)) > 1) or gcd4scc(G) > 1:
        G = num2CG(np.random.randint(2 ** (n ** 2)), n)
    return G


def SM_fixed(Gstar, G, iter=5):
    """ 
    (Ask)

    :param Gstar:
    :type Gstar:
    
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :param iter:
    :type iter: integer
    
    :returns: 
    :rtype: 
    """
    compat = []
    for j in range(0, iter):
        if Gstar == undersample(G, j):
            compat.append(j)
    return compat


def SM_converging(Gstar, G):
    """
    Gstar is the undersampled reference graph, while G is the starting
    graph. The  code searches  over all undersampled  version of  G to
    find all matches with Gstar
    
    :param Gstar:
    :type Gstar:
    
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :returns: 
    :rtype: 
    """
    compat = []
    GG = G
    Gprev = G
    if G == Gstar:
        return [0]
    j = 1
    G = undersample(GG, j)
    while not (G == Gprev):
        if Gstar == G:
            compat.append(j)
        j += 1
        Gprev = G
        G = undersample(GG, j)
    return compat


def search_match(Gstar, G, iter=5):
    """ 
    :param Gstar:
    :type Gstar:
    
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :param iter:
    :type iter: integer
    
    :returns: 
    :rtype: 
    """
    if gcd4scc(G) > 1:
        return SM_fixed(Gstar, G, iter=iter)
    return SM_converging(Gstar, G)


def has_sink(G):
    """ 
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :returns: 
    :rtype: 
    """
    return not reduce(operator.and_, [bool(G[n]) for n in G], True)


def has_root(G):
    """ 
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :returns: 
    :rtype: 
    """
    return has_sink(ecj.tr(G))


def gcd4scc(SCC):
    """ 
    Returns the greatest common divisor of simple loop lengths and in one SCC

    :param SCC: ``gunfolds`` graph
    :type SCC: dictionary (``gunfolds`` graphs)
    
    :returns: the greatest common divisor of simple loop lengths and in one SCC
    :rtype: integer
    """
    g = graph2nx(SCC)
    return ecj.listgcd([len(x) for x in nx.simple_cycles(g)])


def compatible_at_u(uGstar):
    """ 
    :param uGstar:
    :type uGstar:
    
    :returns: 
    :rtype: 
    """
    compat = []
    n = len(uGstar)
    numG = 2 ** (n ** 2)
    # pbar = Percentage()
    for i in range(1, numG):
        G = num2CG(i, n)
        # pbar.update(i+1)
        if len(ecj.scc(G)) > 1:
            continue
        l = search_match(uGstar, G, iter=5)
        if l:
            compat.append((l, G))
    # pbar.finish()
    return compat



import sys
from gunfolds.utils.bfutils import undersample
from gunfolds.utils import ecj
from gunfolds.utils import load_data
import igraph
import math
import numpy as np
import io as StringIO


def graph2dict(g):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :returns: 
    :rtype: 
    """
    D = {}
    for v in range(0, len(g.vs)):
        D[g.vs[v]["label"]] = {}
        for u in g.neighbors(v, mode="OUT"):
            D[g.vs[v]["label"]][g.vs[u]["label"]] = 1
    return D


def paintSCC(g, cm):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)
    
    :param cm:
    :type cm:
    """
    D = graph2dict(g)
    scc = ecj.scc(D)
    for i in range(0, len(scc)):
        for v in scc[i]:
            g.vs[g.vs['label'].index(v)]["color"] = cm[i]


fstr = "%.5f"  # precision to use


def dict2graph(D):
    """
    :param D:
    :type D: dictionary
    
    :returns: ``gunfolds`` graph
    :rtype: dictionary (``gunfolds`` graphs)
    """
    A = np.zeros([len(D), len(D)])
    nodes = []
    indices = {}
    c = 0
    for v in D:
        nodes.append(v)
    idx = np.argsort([int(v) for v in nodes])
    nodes = [nodes[i] for i in idx]
    for v in nodes:
        indices[v] = c
        c += 1
    for v in nodes:
        for u in D[v]:
            if D[v][u] not in (2, 3):
                A[indices[v], indices[u]] = 1
    g = igraph.Graph.Adjacency(A.tolist())
    g.vs["label"] = nodes
    return g


def matrix_start(mname='generic', w_gap=0.45, h_gap=0.5, stl=''):
    print("\\matrix ("
          + mname
          + ") [matrix of nodes, row sep="
          + str(h_gap)
          + "cm,column sep="
          + str(w_gap)
          + "cm"
          + stl
          + "]")
    print("{")


def matrix_end():
    print("};")


def matrix_grid(G, s, mname='generic', w_gap=0.5, h_gap=0.5, type="obs", stl=''):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :param s:
    :type s: (guess) integer
    
    :param mname:
    :type mname: (guess) string
    
    :param w_gap:
    :type w_gap: (guess) float

    :param h_gap:
    :type h_gap: (guess) float

    :param type:
    :type type: (guess) string
    
    :param stl:
    :type stl: (guess) string
    """
    matrix_start(mname=mname, w_gap=w_gap, h_gap=h_gap, stl=stl)
    keylist = G.keys()
    idx = np.argsort([int(v) for v in keylist])
    keylist = [keylist[i] for i in idx]
    for v in keylist:
        print(" ", "& ".join(["\\node[" + type + "]{" + v + "};" for i in range(1, s)]) + "\\\\")
    matrix_end()


def matrix_edges(G, s, mname='generic'):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :param s:
    :type s: (guess) integer
    
    :param mname:
    :type mname: (guess) string
    """
    nodes = list(G.keys())
    idx = np.argsort([int(v) for v in nodes])
    nodes = [nodes[i] for i in idx]
    for v in nodes:
        idx1 = nodes.index(v) + 1
        for u in G[v]:
            if G[v][u] in (1, 3):
                idx2 = nodes.index(u) + 1
                print('\\foreach \\x in{' + ','.join(map(str, range(1, s - 1))) + '}{')
                print('  \\pgfmathtruncatemacro{\\xn}{\\x+1}')
                print('  \\draw[pil] (' + mname + '-' + str(idx1) + '-\\x) -- ('
                      + mname + '-' + str(idx2) + '-\\xn);')
                print('};')


def emacs_vars():
    print('%%% Local Variables:')
    print('%%% mode: latex')
    print('%%% TeX-master: "../master"')
    print('%%% End:')


def dbnprint(G, s, mname='generic', w_gap=0.5, h_gap=0.5, type="obs", stl=''):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :param s:
    :type s: (guess) integer
    
    :param mname:
    :type mname: (guess) string
    
    :param w_gap:
    :type w_gap: (guess) float

    :param h_gap:
    :type h_gap: (guess) float

    :param type:
    :type type: (guess) string
    
    :param stl:
    :type stl: (guess) string
    """
    matrix_grid(G, s, mname=mname, w_gap=w_gap, h_gap=h_gap, type=type, stl=stl)
    matrix_edges(G, s, mname=mname)


def getangle(A, B):
    """
    When A and  B are two angles around the clock  returns an angle of
    the line that is connecting them.
    
    :param A:
    :type A:
    
    :param B:
    :type B:
    
    :returns: returns an angle of the line that is connecting them.
    :rtype: 
    """
    x = np.array([np.cos(np.deg2rad(A)), np.sin(np.deg2rad(A))])
    y = np.array([np.cos(np.deg2rad(B)), np.sin(np.deg2rad(B))])
    d = y - x
    return np.rad2deg(math.atan2(d[1], d[0]))


def cdbnprint(G, mtype="obs", bend=5, curve=5, R=1):
    """
    Prints  out  a  compressed  dbn  repesentation  of  the  graph  in
    TikZ/Latex format

    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :param mtype:
    :type mtype: (guess) string
    
    :param bend:
    :type bend: (guess) integer
    
    :param curve:
    :type curve: (guess) integer
    
    :param R:
    :type R: (guess) integer
    
    :returns: 
    :rtype: 
    """
    output = StringIO.StringIO()
    BE = set()
    n = len(G)
    nodes = list(G.keys())
    idx = np.argsort([int(v) for v in nodes])
    nodes = [nodes[i] for i in idx]

    g = dict2graph(ecj.cloneBfree(G))
    paintSCC(g, load_data.colors)

    for i in range(0, n):
        node = g.vs[i]['label']
        rc = g.vs[i]["color"]
        print("{ \\definecolor{mycolor}{RGB}{" + str(rc[0]) + "," + str(rc[1]) + "," + str(rc[2]) + "}", file=output)
        print("\\node[" + mtype + ", fill=mycolor] (" + str(node) + ") at (" +
              str(-i * 360 / n + 180) + ":" + str(R) + ") {" + str(node) + "};}",
              file=output)

#    print >>output,"\\foreach \\name/\\angle in {"+",".join(
#        [nodes[i]+"/"+str(-i*360/n+180) for i in range(0,n)])+"}"
#    print >>output,"\\node["+mtype+"] (\\name) at (\\angle:"\
#        +str(R)+") {\\name};"

    for i in range(0, n):
        v = nodes[i]
        ll = [(v, u) for u in G[v]]
        for l in ll:
            a, b = l
            if G[a][b] in (2, 3):
                if not(BE.intersection([(a, b)]) or BE.intersection([(b, a)])):
                    print('  \\draw[pilip, on layer=back] (' + str(a) + ') -- (' + str(b) + ');', file=output)
            if G[a][b] in (1, 3):
                ang_a = -nodes.index(a) * 360 / n + 180
                ang_b = -nodes.index(b) * 360 / n + 180
                if a == b:
                    print("\\path[overlay,draw,pil] (" + str(a) + ")" +
                          " .. controls +(" + "%.5f" % (bend + ang_a) +
                          ":" + fstr % (2 * curve) + "mm) and +(" +
                          "%.5f" % (ang_a - bend) +
                          ":" + "%.5f" % (2 * curve) + "mm) .. (" + str(b) + ");", file=output)
                else:
                    print("\\path[overlay,draw,pil] (" + str(a) + ")" +
                          " .. controls +(" + "%.5f" % (bend + getangle(ang_a, ang_b)) +
                          ":" + fstr % curve + "mm) and +(" +
                          fstr % (getangle(ang_b, ang_a) - bend) +
                          ":" + fstr % curve + "mm) .. (" + str(b) + ");", file=output)
    return output


def gprint(G, mtype="obs", bend=5, curve=5, R=1, layout=None):
    """
    Prints out an automatically layout compressed dbn repesentation of
    the graph in TikZ/Latex format

    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :param mtype:
    :type mtype: (guess) string
    
    :param bend:
    :type bend: (guess) integer
    
    :param curve:
    :type curve: (guess) integer
    
    :param R:
    :type R: (guess) integer
    
    :param layout:
    :type layout:
    
    :returns: 
    :rtype: 
    """
    output = StringIO.StringIO()
    BE = set()
    n = len(G)
    if not layout:
        g = dict2graph(ecj.cloneBfree(G))
        layout = g.layout_fruchterman_reingold(maxiter=50000, coolexp=1.1)
        # layout = g.layout_graphopt(niter=50000, node_charge=0.08)
        layout.center([0, 0])
        layout.scale(float(1 / np.absolute(layout.coords).max()))
        layout.scale(R)
        cc = np.round_(np.array(layout.coords), decimals=4)
    else:
        g = dict2graph(ecj.cloneBfree(G))
        cc = np.array(layout.coords)
    paintSCC(g, load_data.colors)
    for i in range(0, n):
        node = g.vs[i]['label']
        rc = g.vs[i]["color"]
        print("{ \\definecolor{mycolor}{RGB}{"
              + str(rc[0]) + "," + str(rc[1]) + "," + str(rc[2]) + "}",
              file=output)
        mcolor = "fill = {rgb: red," + str(rc[0]) + "; green," + str(rc[1]) +\
            "; blue," + str(rc[2]) + "}"
        print("\\node[" + mtype + ", fill=mycolor] (" + node + ") at (" +
              str(cc[i][0]) + "," + str(cc[i][1]) + ") {" + node + "};}",
              file=output)

    for i in range(0, n):
        v = g.vs[i]['label']
        ll = [str(v) + '/' + str(u) for u in G[v]]
        for l in ll:
            a, b = l.split('/')
            if G[a][b] in (2, 3):
                if not(BE.intersection([(a, b)]) or BE.intersection([(b, a)])):
                    print('  \\draw[pilip, on layer=back] (' +
                          a + ') -- (' + b + ');', file=output)
            if G[a][b] in (1, 3):
                if a == b:
                    dff = cc[g.vs['label'].index(a)] - np.mean(cc, 0)
                    ang = np.arctan2(dff[1], dff[0])
                    ang_a = np.rad2deg(ang)
                    print("\\path[overlay,draw,pil] (" + a + ")" +
                          " .. controls +(" + "%.5f" % (bend + ang_a) +
                          ":" + fstr % (2 * curve) + "mm) and +(" +
                          "%.5f" % (ang_a - bend) +
                          ":" + "%.5f" % (2 * curve) + "mm) .. (" + b + ");",
                          file=output)
                else:
                    dff = cc[g.vs['label'].index(b)] \
                        - cc[g.vs['label'].index(a)]
                    ang = np.arctan2(dff[1], dff[0])
                    ang_a = np.rad2deg(ang)
                    ang_b = ang_a + 180
                    print("\\path[overlay,draw,pil] (" + a + ")" +
                          " .. controls +(" +
                          "%.5f" % (bend + ang_a) +
                          ":" + fstr % curve + "mm) and +(" +
                          fstr % (ang_b - bend) +
                          ":" + fstr % curve + "mm) .. (" + b + ");",
                          file=output)
    return output


def cdbnwrap(G, u, name='AAA', R=1, gap=0.5):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :param u: (guess) maximum under sampling rate
    :type u: (guess) integer
    
    :param name:
    :type name: (guess) string
    
    :param R:
    :type R: (guess) integer
    
    :param gap:
    :type gap: (guess) float
    
    :returns: 
    :rtype: 
    """
    output = StringIO.StringIO()
    print("\\node[right=" + str(gap) + "cm of " + name + str(u - 1)
          + ",scale=0.7](" + name + str(u) + "){", file=output)
    print("\\begin{tikzpicture}",  file=output)
    s = cdbnprint(undersample(G, u), mtype='lahid', bend=25, curve=10, R=R)
    print(s.getvalue(),  file=output)
    s.close()
    print("\\end{tikzpicture}",  file=output)
    print("};",  file=output)

    return output


def cdbnsingle(g, scale=0.7, R=1, mtype="lahid"):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)

    :param scale:
    :type scale: (guess) float
    
    :param R:
    :type R: (guess) integer
    
    :param mtype:
    :type mtype: (guess) string
    
    :returns: 
    :rtype: 
    """
    output = StringIO.StringIO()
    print("\\node[scale=" + str(scale) + "](){", file=output)
    print("\\begin{tikzpicture}", file=output)
    s = cdbnprint(g, mtype=mtype, bend=25, curve=10, R=R)
    print(s.getvalue(), file=output)
    s.close()
    print("\\end{tikzpicture}", file=output)
    print("};", file=output)
    return output


def cdbn_single(G, u, scale=0.7, R=1, gap=0.5, mtype="lahid"):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)
    
    :param u: (guess) maximum under sampling rate
    :type u: (guess) integer
    
    :param scale:
    :type scale: (guess) float
    
    :param R:
    :type R: (guess) integer
    
    :param gap:
    :type gap: (guess) float
    
    :param mtype:
    :type mtype: (guess) string
    
    :returns: 
    :rtype: 
    """
    return cdbnsingle(undersample(G, u), scale=scale, R=R, gap=gap, mtype=mtype)


def gsingle(g, scale=0.7, R=1, mtype="lahid", layout=None):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)

    :param scale:
    :type scale: (guess) float
    
    :param R:
    :type R: (guess) integer
    
    :param mtype:
    :type mtype: (guess) string
    
    :param layout:
    :type layout:
    
    :returns: 
    :rtype: 
    """
    output = StringIO.StringIO()
    print("\\node[scale=" + str(scale) + "](){", file=output)
    print("\\begin{tikzpicture}", file=output)
    s = gprint(g, mtype=mtype, bend=25, curve=6, R=R, layout=layout)
    print(s.getvalue(), file=output)
    s.close()
    print("\\end{tikzpicture}", file=output)
    print("};", file=output)
    return output


def g_single(G, u, scale=0.7, R=1, gap=0.5, mtype="lahid", layout=None):
    """
    :param G: ``gunfolds`` format graph
    :type G:  dictionary (``gunfolds`` graphs)
    
    :param u: maximum under sampling rate
    :type u: integer
    
    :param scale:
    :type scale: (guess) float
    
    :param R:
    :type R: (guess) integer
    
    :param gap:
    :type gap: (guess) float
    
    :param mtype:
    :type mtype: (guess) string
    
    :param layout:
    :type layout:
    
    :returns: 
    :rtype: 
    """
    g = undersample(G, u)
    return gsingle(g, scale=scale, R=R, gap=gap, mtype=mtype, layout=layout)


def unfoldplot(G, steps=7, repeats=5, gap=0.5, R=1, hg=0.1, wgap=0.7, name='AAA', stl=''):
    """
    :param G: ``gunfolds`` format graph
    :type G:  dictionary (``gunfolds`` graphs)
    
    :param steps:
    :type steps: (guess) integer
    
    :param repeats:
    :type repeats: (guess) integer
    
    :param gap:
    :type gap: (guess) float
    
    :param R:
    :type R: (guess) integer
    
    :param hg:
    :type hg: (guess) float
    
    :param wgap:
    :type wgap: (guess) float
    
    :param name:
    :type name: (guess) string
    
    :param stl:
    :type stl: (guess) string
    """
    u = 0
    dbnprint(undersample(G, u), repeats, w_gap=wgap,
             h_gap=hg, mname=name + str(u), type='hid', stl=stl)
    print("\\node[left=" + str(gap) + "cm of " + name + str(u) + ",scale=0.7] (C) {")
    print("\\begin{tikzpicture}")
    cdbnprint(G, mtype='hid', bend=15, curve=5, R=R)
    print("\\end{tikzpicture}")
    print("};")
    for u in range(1, steps):
        dbnprint(undersample(G, u), repeats, w_gap=wgap, h_gap=hg, mname=name +
                 str(u), type='ahid', stl=', below=0.25cm of ' + name + str(u - 1))

        print("\\node[left=" + str(gap) + "cm of " + name + str(u) + ",scale=0.7] () {")
        print("\\begin{tikzpicture}")
        cdbnprint(undersample(G, u), mtype='lahid', bend=15, curve=5, R=R)
        print("\\end{tikzpicture}")
        print("};")
    emacs_vars()


def foldplot(G, steps=7, gap=0.5, R=1, name='AAA', stl='', bend=50, curve=10):
    """
    :param G: ``gunfolds`` format graph
    :type G:  dictionary (``gunfolds`` graphs)
    
    :param steps:
    :type steps: (guess) integer
    
    :param gap:
    :type gap: (guess) float
    
    :param R:
    :type R: (guess) integer
    
    :param name:
    :type name: (guess) string
    
    :param stl:
    :type stl: (guess) string

    :param bend:
    :type bend:

    :param curve:
    :type curve:
    """
    u = 0
    print('\\node[scale=0.7' + stl + '] (' + name + str(u) + ') {')
    print("\\begin{tikzpicture}")
    s = cdbnprint(G, mtype='hid', bend=bend, curve=curve, R=R)
    print(s.getvalue())
    s.close()
    print("\\end{tikzpicture}")
    print("};")
    for u in range(1, steps):
        s = cdbnwrap(G, u, R=R, gap=gap, name=name)
        print(s.getvalue())
        s.close()
    emacs_vars()


def matrix_fold(G, m, n, R=1, mname='g', w_gap=0.5, h_gap=0.5, stl=''):
    """
    :param G: ``gunfolds`` format graph
    :type G:  dictionary (``gunfolds`` graphs)
    
    :param m:
    :type m:
    
    :param n:
    :type n:
    
    :param R:
    :type R:
    
    :param mname:
    :type mname:
    
    :param w_gap:
    :type w_gap:
    
    :param h_gap:
    :type h_gap:
    
    :param stl:
    :type stl:
    """
    matrix_start(mname=mname, w_gap=w_gap, h_gap=h_gap, stl=stl)
    for i in range(0, n):
        S = []
        for j in range(0, m):
            u = m * i + j
            if u == 0:
                s = cdbn_single(G, u, R=R, mtype="hid")
            else:
                s = cdbn_single(G, u, R=R)
            S.append(s.getvalue())
            s.close()
        print(" ", "& ".join(S) + "\\\\")
    matrix_end()


def gmatrix_fold(G, m, n, R=1, mname='g', w_gap=0.5, h_gap=0.5, stl='', shift=0):
    # g = dict2graph(ecj.cloneBfree(G))
    # layout = g.layout_graphopt(niter=50000, node_charge=0.02)
    # layout.center([0,0])
    # layout.scale(0.01)
    """
    :param G: ``gunfolds`` format graph
    :type G:  dictionary (``gunfolds`` graphs)
    
    :param m:
    :type m:
    
    :param n:
    :type n:
    
    :param R:
    :type R:
    
    :param mname:
    :type mname:
    
    :param w_gap:
    :type w_gap:
    
    :param h_gap:
    :type h_gap:
    
    :param stl:
    :type stl:
    
    :param shift:
    :type shift:
    """
    matrix_start(mname=mname, w_gap=w_gap, h_gap=h_gap, stl=stl)
    for i in range(0, n):
        S = []
        for j in range(0, m):
            u = m * i + j + shift
            s = g_single(G, u, R=R, mtype="hid")
            S.append(s.getvalue())
            s.close()
        print(" ", "& ".join(S) + "\\\\")
    matrix_end()


def addselfedge(G, v):
    """
    :param G: ``gunfolds`` format graph
    :type G:  dictionary (``gunfolds`` graphs)
    
    :param v:
    :type v:
    """
    G[v].update({v: 1})


def gmatrix_list(l, m, n, R=1, mname='g', w_gap=0.5, h_gap=0.5, stl='', shift=0):
    """
    Given a list of graphs prints them out as latex laying out each graph in a force-based layout
    
    :param l:
    :type l:
    
    :param m:
    :type m:
    
    :param n:
    :type n:
    
    :param R:
    :type R:
    
    :param mname:
    :type mname:
    
    :param w_gap:
    :type w_gap:
    
    :param h_gap:
    :type h_gap:
    
    :param stl:
    :type stl:
    
    :param shift:
    :type shift:
    """
    matrix_start(mname=mname, w_gap=w_gap, h_gap=h_gap, stl=stl)
    for i in range(0, n):
        S = []
        for j in range(0, m):
            u = m * i + j + shift
            if u > len(l) - 1:
                break
            s = gsingle(l[u], R=R, mtype="hid")
            S.append(s.getvalue())
            s.close()
        print(" ", "& ".join(S) + "\\\\")
    matrix_end()


def matrix_list(l, m, n, R=1, mname='g', w_gap=0.5, h_gap=0.5, stl='', shift=0):
    """
    Given a list of graphs prints them out as latex laying out each in a circle layout
    
    :param l:
    :type l:
    
    :param m:
    :type m:
    
    :param n:
    :type n:
    
    :param R:
    :type R:
    
    :param mname:
    :type mname:
    
    :param w_gap:
    :type w_gap:
    
    :param h_gap:
    :type h_gap:
    
    :param stl:
    :type stl:
    
    :param shift:
    :type shift:
    """
    matrix_start(mname=mname, w_gap=w_gap, h_gap=h_gap, stl=stl)
    for i in range(0, n):
        S = []
        for j in range(0, m):
            u = m * i + j + shift
            if u > len(l) - 1:
                break
            s = cdbnsingle(l[u], R=R, mtype="hid")
            S.append(s.getvalue())
            s.close()
        print(" ", "& ".join(S) + "\\\\")
    matrix_end()


class WritableObject:

    def __init__(self):
        self.content = []

    def write(self, string):
        self.content.append(string)


def output_graph_figure(g):
    """
        Given a graph will out put a latex file of the graph visualization to figures directory
        
        :param g: ``gunfolds`` graph
        :type g:  dictionary (``gunfolds`` graphs)
    """
    sys.stdout = open("../figures/shipfig_figure.tex", "w")
    foldplot(g, steps=1, gap=5, R=15)
    sys.stdout.close()
    sys.stdout = sys.__stdout__



import graph_tool as gt
from graph_tool import draw as gtd
from graph_tool.draw.cairo_draw import position_parallel_edges
import gunfolds.utils.graphkit as gk
import gunfolds.utils.bfutils as bfu
import numpy as np
from gi.repository import Gdk
from gi import require_version
require_version("Gdk", "3.0")


def getscreensize():
    """
    Returns a tuple ``(w,h)`` with the largest width ``w`` and height ``h`` among all connected monitors.

    :returns: a tuple ``(w,h)``
    :rtype: a tuple with integer values
    """
    # from here https://stackoverflow.com/a/65894179
    screen = Gdk.Display.get_default()
    w = 0
    h = 0
    for x in range(0, screen.get_n_monitors()):
        xw = screen.get_monitor(x).get_geometry().width
        xh = screen.get_monitor(x).get_geometry().height
        if w < xw:
            w = xw
        if h < xh:
            h = xh
    return w, h


def circ_position(gg):
    """
    Generates node positions for graph-tool graph that arrange them in a ring

    :param gg: ``graph-tool`` graph object
    :type gg: a ``graph-tool`` object of class Graph
    """
    pos = gg.new_vertex_property("vector<double>")
    n = gg.num_vertices()
    s = 2.0*np.pi/n
    for v in range(n):
        idx = int(gg.vertex_properties['label'][gg.vertex(v)]) - 1
        pos[gg.vertex(v)] = (n * np.cos(s * idx),
                             n * np.sin(s * idx))
    gg.vertex_properties["pos"] = pos


def every_edge_control_points(n):
    """
    Generate a graph-tool Graph class object with all possible edges (a superclicue).

    The function generates a graph and populates all parameters for all edges.
    It first does not include the bidirected adges to compute the control_points for
    directed edges so that they all curve. Subsequently it adds bidirected (red) edges
    and resets their control points to have all bidirected edges be rendered as straight lines.

    :param n: number of nodes
    :type n: integer

    :returns: ``graph-tool`` graph
    :rtype: a ``graph-tool`` object of class Graph
    """

    g = gk.fullyconnected(n)

    gr = gt.Graph()

    vlabel = gr.new_vertex_property("string")
    ecolor = gr.new_edge_property("string")
    ep_width = gr.new_edge_property("float")
    em_size = gr.new_edge_property("float")

    verts = {}
    edges = {}

    for v in g:
        verts[v] = gr.add_vertex()
        vlabel[verts[v]] = str(v)
    gr.vertex_properties["label"] = vlabel
    circ_position(gr)

    for v in g:
        for w in g[v]:
            edges[(v, w)] = gr.add_edge(verts[v], verts[w])
            ecolor[edges[(v, w)]] = 'k'
            ep_width[edges[(v, w)]] = 1
            em_size[edges[(v, w)]] = 15

    control = position_parallel_edges(gr, gr.vertex_properties["pos"])

    for v in range(1, n+1):
        for w in range(v, n+1):
            edges[(v, w)] = gr.add_edge(verts[v], verts[w])
            ecolor[edges[(v, w)]] = 'r'
            ep_width[edges[(v, w)]] = 1
            em_size[edges[(v, w)]] = 1
            control[edges[(v, w)]] = []

    gr.edge_properties["color"] = ecolor
    gr.edge_properties["pen_width"] = ep_width
    gr.edge_properties["marker_size"] = em_size
    gr.edge_properties["control"] = control

    return gr


def colorcomponents(gr):
    """
    Assigns a color to each vertex to ensure that vertices from the same strongly connected component
    have the same color.

    By default 12 distinct colors from colorbrewer2 quantitative palette are used and anything beyond
    that is assigned a white color. In the future that behaviour may change to assigning random colors
    or to assigning useful semantics to color temperature or shade (such as SCC density or size).

    :param gr: ``graph-tool`` graph
    :type gr: a ``graph-tool`` object of class Graph
    """

    colors = ["#9feb3dff",
              '#a6cee3ff', '#1f78b4ff', '#b2df8aff',
              '#33a02cff', '#fb9a99ff', '#e31a1cff',
              '#fdbf6fff', '#ff7f00ff', '#cab2d6ff',
              '#6a3d9aff', '#ffff99ff', '#b15928ff']
    ecolor = gr.edge_properties["color"]
    vcolor = gr.new_vertex_property("string")

    is_true_directed = gr.new_edge_property("bool")
    for edge in gr.edges():
        is_true_directed[edge] = ecolor[edge] == 'k'
    g = gt.GraphView(gr, efilt=is_true_directed)
    comp, hist = gt.topology.label_components(g)
    for i, vertex in enumerate(gr.vertices()):
        try:
            vcolor[vertex] = colors[comp[i]]
        except IndexError:
            vcolor[vertex] = '#00000000'
    gr.vertex_properties["color"] = vcolor


def g2gt(g):
    """
    Converts a ``gunfolds`` graph to an object of class Graph of ``graph-tool`` package.
    This includes setting all parameters of edges and vertices but vertex colors.

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)

    :returns: ``graph-tool`` graph
    :rtype: a ``graph-tool`` object of class Graph
    """

    gr = every_edge_control_points(len(g))
    color = gr.edge_properties["color"]

    edge_removal = []
    for edge in gr.edges():
        v = gr.vertex_index[edge.source()]+1
        w = gr.vertex_index[edge.target()]+1
        if color[edge] == 'r' and w in g[v] and g[v][w] in {2, 3}:
            continue
        if color[edge] == 'k' and w in g[v] and g[v][w] in {1, 3}:
            continue
        edge_removal.append(edge)

    for edge in edge_removal:
        gr.remove_edge(edge)

    return gr


def gt2g(gr, dir_c='k', bidir_c='r'):
    """
    Converts a ``graph-tool`` object of class Graph to ``gunfolds`` graph

    :param gr: ``gunfolds`` graph
    :type gr: dictionary(``gunfolds`` graph)

    :returns: ``gunfolds`` graph
    :rtype: dictionary (``gunfolds`` graphs)
    """
    # check if the graph-tool graph has edge colors
    try:
        color = gr.edge_properties["color"]
    except KeyError:
        color = []

    g = {}
    for v in gr.vertices():
        g[int(str(v))+1] = {}
        if v.out_edges():
            if color:
                for x in v.out_edges():
                    if color[x] == dir_c:
                        g[int(str(v))+1][int(str(x.target()))+1] = 1
                for x in v.out_edges():
                    if color[x] == bidir_c:
                        if int(str(x.target()))+1 in g[int(str(v))+1]:
                            g[int(str(v))+1][int(str(x.target()))+1] = 3
                        else:
                            g[int(str(v))+1][int(str(x.target()))+1] = 2
                for x in v.in_edges():
                    if color[x] == bidir_c:
                        if int(str(x.source()))+1 in g[int(str(v))+1]:
                            g[int(str(v))+1][int(str(x.source()))+1] = 3
                        else:
                            g[int(str(v))+1][int(str(x.source()))+1] = 2
            else:
                g[int(str(v))+1] = {int(str(x.target()))+1: 1 for x in v.out_edges()}
    return g


def hshift(g, shift):
    """
    Horizontally shift positions of all nodes of ``graph-tool`` graph ``g`` by the value of ``shift``

    :param g: ``graph-tool`` graph
    :type g: a ``graph-tool`` object of class Graph

    :param shift: shift value
    :type shift: float
    """
    pos = g.vertex_properties["pos"]
    for x in g.vertices():
        pos[x][0] = pos[x][0] + shift


def linegraph(glist, sccs=True):
    """
    Takes a list of ``gunfolds`` graphs and merges them into a single
    Graph class ``graph-tool`` object taking care of node positions of
    each graph to ensure that the constituent graphs are arranged in a single raw when plotted.

    :param glist: a list of ``gunfold`` graphs
    :type glist: list of dictionaries (``gunfolds`` graphs)

    :param sccs: whether to distinguish SCCs by color
    :type sccs: boolean

    :returns: ``graph-tool`` graph
    :rtype: a ``graph-tool`` object of class Graph
    """

    gr = gt.Graph()

    shift = 0
    for i, g in enumerate(glist):
        g_i = g2gt(g)
        if sccs:
            colorcomponents(g_i)
        hshift(g_i, i*shift)
        points = np.asarray([x for x in g_i.vertex_properties["pos"]])
        shift = 1.5*(points[:, 0].max() - points[:, 0].min())
        gt.generation.graph_union(gr, g_i,
                                  internal_props=True,
                                  include=True)

    return gr


def plotg(g, sccs=True, output=None, fmt='auto'):
    """
    Given a ``gunfolds`` graph, plots it in an interactive window

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)

    :param sccs: whether to distinguish SCCs by color
    :type sccs: boolean

    :param output: Output file name (or object). If not given, the graph will be displayed via interactive_window().
    :type output: string or file object (optional, default: None)

    :param fmt: Output file format. Possible values are ``"auto"``, ``"ps"``, ``"pdf"``, ``"svg"``, and ``"png"``. If the value is ``"auto"``, the format is guessed from the output parameter.
    :type fmt: string or file object (optional, default: None)
    """

    gg = g2gt(g)

    vcolors = [0.62109375, 0.875, 0.23828125, 1]
    if sccs:
        colorcomponents(gg)
        vcolors = gg.vertex_properties["color"]

    gtd.graph_draw(gg, gg.vertex_properties["pos"],
                   vertex_text=gg.vertex_properties['label'],
                   edge_pen_width=gg.edge_properties['pen_width'],
                   edge_marker_size=gg.edge_properties['marker_size'],
                   edge_control_points=gg.edge_properties['control'],
                   vertex_pen_width=1,
                   edge_color=gg.edge_properties['color'],
                   vertex_fill_color=vcolors, output=output, fmt=fmt)


def plotgunfolds(g, sccs=True, output=None, fmt='auto'):
    """
    Given a ``gunfolds`` graph plots all of its undersamples versions arranged sequentially in a horizontal line.

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)

    :param interactive: whether to make the window interactive
    :type interactive: boolean

    :param sccs: whether to distinguish SCCs by color
    :type sccs: boolean

    :param output: Output file name (or object). If not given, the graph will be displayed via interactive_window().
    :type output: string or file object (optional, default: None)

    :param fmt: Output file format. Possible values are ``"auto"``, ``"ps"``, ``"pdf"``, ``"svg"``, and ``"png"``. If the value is ``"auto"``, the format is guessed from the output parameter.
    :type fmt: string or file object (optional, default: None)
    """

    x = bfu.all_undersamples(g)
    gg = linegraph(x, sccs=sccs)

    points = np.asarray([x for x in gg.vertex_properties["pos"]])
    width = points[:, 0].max() - points[:, 0].min()
    height = points[:, 1].max() - points[:, 1].min()
    f_w = getscreensize()[0]
    f_h = height/width * f_w
    hshift(gg, -width)

    vcolors = [0.62109375, 0.875, 0.23828125, 1]
    if sccs:
        vcolors = gg.vertex_properties["color"]

    gtd.graph_draw(gg, gg.vertex_properties["pos"],
                   vertex_text=gg.vertex_properties['label'],
                   edge_pen_width=gg.edge_properties['pen_width'],
                   edge_marker_size=gg.edge_properties['marker_size'],
                   edge_control_points=gg.edge_properties['control'],
                   vertex_pen_width=1,
                   adjust_aspect=True,
                   fit_view=True,
                   fit_view_ink=True,
                   output_size=(f_w, f_h),
                   edge_color=gg.edge_properties['color'],
                   vertex_fill_color=vcolors,
                   output=output, fmt=fmt
                   )



from gunfolds.conversions import g2num, ug2num, num2CG
import gunfolds.utils.zickle as zkl
import gunfolds.utils.graphkit as gk
import itertools
import numpy as np


# Undersampling functions
def is_sclique(G):
    """
    Tests if the graph is a superclique (all possible connections)

    :param G: ``gunfolds`` format graph
    :type G:  dictionary (``gunfolds`` graphs)

    :returns: True, if ``G`` is a super clique
    :rtype: boolean
    """
    n = len(G)
    row_sum = 3 * (n-1) + 1    # rows will have all connections plus a self loop
    for vertex in G:
        if G[vertex].get(vertex) != 1 or sum(G[vertex].values()) != row_sum:
            return False
    return True


def directed_inc(G, D):
    """ 
    A helper function for determining directed edges in an undersampled
    graph given a previously undersampled graph 
        
    :param G: ``gunfolds`` format graph G_1 - without undersampling
    :type G:  dictionary (``gunfolds`` graphs)
    
    :param D: ``gunfolds`` graph G_u a version of G_1 undersampled by u
    :type D:  dictionary (``gunfolds`` graphs)

    :returns:
    :rtype: dictionary (``gunfolds`` graphs)
    """
    G_un = {}
    # directed edges
    for vert1 in D:
        G_un[vert1] = {}
        for vert2 in D[vert1]:
            if G[vert2] and D[vert1][vert2] in (1, 3):
                for e in G[vert2]:
                    G_un[vert1][e] = 1
    return G_un


def bidirected_inc(G, D):
    """ 
    A helper function for determining bidirected edges in an undersampled
    graph given a previously undersampled graph 
    
    :param G: ``gunfolds`` format graph G_1 - without undersampling
    :type G:  dictionary (``gunfolds`` graphs)
    
    :param D: ``gunfolds`` graph G_u a version of G_1 undersampled by u
    :type D:  dictionary (``gunfolds`` graphs)

    :returns:
    :rtype:
    """
    for vert1 in G:
        # transfer old bidirected edges
        for vert2 in D[vert1]:
            if D[vert1][vert2] in (2, 3):
                G[vert1][vert2] = 2 if G[vert1].get(vert2, 2) == 2 else 3
        # new bidirected edges
        edges = [e for e in D[vert1] if D[vert1][e] in (1, 3)]
        for pair in itertools.permutations(edges, 2):
            G[pair[0]][pair[1]] = 2 if G[pair[0]].get(pair[1], 2) == 2 else 3
    return G


def increment_u(G_star, G_u):
    """ 
    Given graph `G_star` - without undersampling - and its undersampled version `G_u`, generate the `G_{u+1}` graph

    :param G_star: ``gunfolds`` format graph G_1 - without undersampling
    :type G_star: dictionary (``gunfolds`` graphs)
    
    :param G_u: ``gunfolds`` graph G_u a version of G_1 undersampled by u
    :type G_u: dictionary (``gunfolds`` graphs)
    
    :returns: 
    :rtype: 
    """
    # directed edges
    G_un = directed_inc(G_star, G_u)
    # bidirected edges
    G_un = bidirected_inc(G_un, G_u)
    return G_un


def pure_directed_inc(G, D):
    """ 
    :param G: ``gunfolds`` format graph G_1 - without undersampling
    :type G:  dictionary (``gunfolds`` graphs)
    
    :param D: ``gunfolds`` graph G_u a version of G_1 undersampled by u
    :type D:  dictionary (``gunfolds`` graphs)
    
    :returns: 
    :rtype: 
    """
    G_un = {}
    # directed edges
    for vert1 in D:
        G_un[vert1] = {}
        for prev_vert in D[vert1]:
            for vert2 in G[prev_vert]:
                G_un[vert1][vert2] = 1
    return G_un


def increment(G):
    """
    Undersample ``G`` by 2
    only works for G1 to G2 directed

    :param G: ``gunfolds`` format graph
    :type G:  dictionary (``gunfolds`` graphs)

    :returns: undersampled graph
    :rtype: dictionary (``gunfolds`` graphs)
    """
    G2 = {n: {} for n in G}

    for vert1 in G:
        for h in G[vert1]:
            for e in G[h]:
                if not (e in G2[vert1]):
                    G2[vert1][e] = 1

    for vert1 in G:
        for pair in itertools.combinations(G[vert1], 2):

            if pair[1] in G2[pair[0]]:
                G2[pair[0]][pair[1]] |= 2
            else:
                G2[pair[0]][pair[1]] = 2

            if pair[0] in G2[pair[1]]:
                G2[pair[1]][pair[0]] |= 2
            else:
                G2[pair[1]][pair[0]] = 2

    return G2


def dincrement_u(G_star, G_u):
    """
    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)

    :param G_u:
    :type G_u:

    :returns:
    :rtype:
    """
    # directed edges
    G_un = pure_directed_inc(G_star, G_u)
    return G_un


def undersample(G, u):
    """
    Undersample graph ``G`` by rate ``u``

    :param G: ``gunfolds`` format graph
    :type G:  dictionary (``gunfolds`` graphs)

    :param u: undersampling rate
    :type u: integer

    :returns: undersampled graph
    :rtype: dictionary (``gunfolds`` graphs)
    """
    if u == 1:
        return G
    Gu = G
    for i in range(u-1):
        Gu = increment_u(G, Gu)
    return Gu


def all_undersamples(G_star):
    """ 
    Returns a list of all undersampled graphs (excluding superclique) 
 
    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)

    :returns: list of all undersampled graphs (excluding superclique)
    :rtype: list of graphs
    """
    glist = [G_star]
    while True:
        g = increment_u(G_star, glist[-1])
        if is_sclique(g):
            return glist  # superclique convergence
        # this will (may be) capture DAGs and oscillations
        if g in glist:
            return glist
        glist.append(g)
    return glist


def call_undersamples(G_star):
    """ 
    Returns a list of all undersampled graphs (including superclique) 
    
    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)

    :returns: list of all undersampled graphs (including superclique)
    :rtype: list of graphs
    """
    glist = [G_star]
    while True:
        g = increment_u(G_star, glist[-1])
        if g in glist:
            return glist
        glist.append(g)
    return glist


def compact_call_undersamples(G_star):
    """ 
    Returns a list of all undersampled graphs (including superclique)
    in binary encoded format 
    
    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)

    :returns: a list of all undersampled graphs (including superclique)
              in binary encoded format
    :rtype: list of integers
    """
    glist = [ug2num(G_star)]
    lastgraph = G_star
    while True:
        g = increment_u(G_star, lastgraph)
        n = ug2num(g)
        if n in glist:
            return glist
        glist.append(n)
        lastgraph = g
    return glist


def cc_undersamples(G_star, steps=1):
    """
    Returns ``G_u`` for ``G_star`` with default value of ``u=2``.
    If the input graph converges before the requested ``u`` - an empty list is returned.

    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)
    
    :param steps: undersampling rate - 1 for G_star
    :type steps: integer

    :returns:
    :rtype:
    """
    glist = [ug2num(G_star)]
    lastgraph = G_star
    for i in range(steps):
        g = increment_u(G_star, lastgraph)
        n = ug2num(g)
        if n in glist:
            return []
        glist.append(n)
        lastgraph = g
    return glist[-1]


# Misc graph functions

def overshoot(G_star, H):
    """
    Undersample ``G_star`` until it turns into a ``G_u`` for which ``H`` is an edge subset.
    Return ``True`` in this case.
    If this does not happen for either of the undersampling rates until teh convergence - return ``False``

    :param G_star: ``gunfolds`` format graph at ``u=1``
    :type G_star: dictionary (``gunfolds`` graphs)

    :param H: ``gunfolds`` format graph at some ``u``
    :type H: dictionary (``gunfolds`` graphs)

    :returns:
    :rtype: boolean
    """
    glist = [G_star]
    while True:
        g = increment_u(G_star, glist[-1])
        if is_sclique(g):
            return False
        if gk.isedgesubset(H, g):
            return True
        if g in glist:
            return False
        glist.append(g)
    return False


def forms_loop(G_star, loop):
    """
    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)

    :param loop:
    :type loop:

    :returns:
    :rtype: boolean
    """
    glist = [G_star]
    while True:
        g = increment_u(G_star, glist[-1])
        if (g2num(gk.digonly(g)) & loop) == loop:
            return True
        if g in glist:
            return False
        glist.append(g)
    return False


def call_u_conflicts_d(G_star, H):
    """
    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)

    :param H:
    :type H:

    :returns:
    :rtype: boolean
    """
    glist = [G_star]
    while True:
        g = dincrement_u(G_star, glist[-1])
        if gk.isedgesubset(g, H):
            return False
        if g in glist:
            return True
        glist.append(g)
    return True


def call_u_conflicts(G_star, H):
    """
    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)

    :param H:
    :type H:

    :returns:
    :rtype: boolean
    """
    glist = [G_star]
    while True:
        # g = increment_u(G_star, glist[-1])
        g = directed_inc(G_star, glist[-1])
        if gk.isdedgesubset(g, H):
            return False
        g = bidirected_inc(g, glist[-1])
        if gk.isedgesubset(g, H):
            return False
        if g in glist:
            return True
        glist.append(g)
    return True


def call_u_conflicts2(G_star, H):
    """
    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)

    :param H:
    :type H:

    :returns:
    :rtype:
    """
    glist = [G_star]
    while True:
        g = increment_u(G_star, glist[-1])
        if gk.isedgesubset(g, H):
            return False, glist
        if g in glist:
            return True, glist
        glist.append(g)
    return True, glist


def call_u_equals2(G_star, glist, H):
    """
    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)

    :param glist: a list of graphs that are undersampled versions of
        the same system
    :type glist: list of dictionaries (``gunfolds`` graphs)

    :param H:
    :type H:

    :returns:
    :rtype: boolean
    """
    while True:
        g = increment_u(G_star, glist[-1])
        if g == H:
            return True
        if g in glist:
            return False
        glist.append(g)
    return False


def call_u_equals(G_star, H):
    """
    :param G_star: ``gunfolds`` format graph
    :type G_star: dictionary (``gunfolds`` graphs)

    :param H:
    :type H:

    :returns:
    :rtype: boolean
    """
    glist = [G_star]
    while True:
        g = increment_u(G_star, glist[-1])
        if g == H:
            return True
        if g in glist:
            return False
        glist.append(g)
    return False


def compatible(d1, d2):
    """
    :param d1:
    :type d1:

    :param d2:
    :type d2:

    :returns:
    :rtype:
    """
    idx = np.where(np.array([[r == l for l in d2] for r in d1]))
    return idx


def compat(G):
    """
    :param G: ``gunfolds`` format graph
    :type G: dictionary (``gunfolds`` graphs)

    :returns:
    :rtype:
    """
    n = len(G)
    # sample all the graph for gStar
    star_l = all_undersamples(G)
    hits = {}
    # brute force all graphs
    for i in range(0, 2 ** (n**2)):
        tmpG = num2CG(i, n)
        tmp_l = all_undersamples(tmpG)
        c = compatible(tmp_l, star_l)
        if len(sum(c)) > 0:
            hits[i] = c
    return hits


def icompat(i, nodes):
    """
    :param i:
    :type i:

    :param nodes:
    :type nodes:

    :returns:
    :rtype:
    """
    print(i)
    g = num2CG(i, nodes)
    return compat(g)


def ilength(i, nodes):
    """
    :param i:
    :type i:

    :param nodes:
    :type nodes:

    :returns:
    :rtype:
    """
    print(i)
    g = num2CG(i, nodes)
    return len(call_undersamples(g))


def iall(i, nodes):
    """
    :param i:
    :type i:

    :param nodes:
    :type nodes:

    :returns:
    :rtype:
    """
    print(i)
    g = num2CG(i, nodes)
    return compact_call_undersamples(g)


def cc_all(i, nodes, steps):
    """
    :param i:
    :type i:

    :param nodes:
    :type nodes:

    :param steps:
    :type steps:

    :returns:
    :rtype:
    """
    g = num2CG(i, nodes)
    return cc_undersamples(g, steps=steps)


def make_rect(l):
    """
    :param l:
    :type l:

    :returns:
    :rtype:
    """
    max_seq = max(map(len, l))
    nl = []
    for e in l:
        e += [e[-1]] * (max_seq - len(e))
        nl.append(e)
    return nl


def loadgraphs(fname):
    """
    Loads a graph from a zickle file

    :param fname: file name
    :type fname: string

    :returns: ``gunfolds`` graph
    :rtype: dictionary (``gunfolds`` graphs)
    """
    g = zkl.load(fname)
    return g


def savegraphs(l, fname):
    """
    Saves a graph into a zickle file format

    :param l: ``gunfolds`` graph
    :type l: dictionary (``gunfolds`` graphs)

    :param fname: file name
    :type fname: string
    """
    zkl.save(l, fname)


# talking about extra edges on top of the ring


def dens2edgenum(d, n=10):
    """
    Convert density into the number of extra edges needed for a ring
    graph to achieve that density

    :param d: density
    :type d: float

    :param n: number of nodes in the graph
    :type n: integer

    :returns: number of extra edges needed for a ring
              graph to achieve that density
    :rtype: integer
    """
    return max(int(d * n**2)-n, 1)


def edgenum2dens(e, n=10):
    """
    Convert number of edges excluding edges of the ring to density

    :param e: number of edges excluding edges of the ring
    :type e: integer

    :param n: number of nodes in the graph
    :type n: integer

    :returns: density
    :rtype: double
    """
    return np.double(e + n)/n**2


def check_conflict(H, G_test, au=None):
    """
    :param H:
    :type H:

    :param G_test:
    :type G_test:

    :param au:
    :type au:

    :returns:
    :rtype:
    """
    if not au:
        allundersamples = call_undersamples(G_test)
    else:
        allundersamples = au
    for graph in allundersamples:
        if gk.isedgesubset(graph, H):
            return False
    return True


def check_conflict_(Hnum, G_test, au=None):
    """
    :param Hnum:
    :type Hnum:

    :param G_test:
    :type G_test:

    :param au:
    :type au:

    :returns:
    :rtype:
    """
    if not au:
        allundersamples = call_undersamples(G_test)
    else:
        allundersamples = au
    # Hnum = ug2num(H)
    for graph in allundersamples:
        gnum = ug2num(graph)
        if gnum[0] & Hnum[0] == gnum[0] and gnum[1] & Hnum[1] == gnum[1]:
            return False
    return True


def check_equality(H, G_test, au=None):
    """
    :param H:
    :type H:

    :param G_test:
    :type G_test:

    :param au:
    :type au:

    :returns:
    :rtype:
    """
    if not au:
        allundersamples = call_undersamples(G_test)
    else:
        allundersamples = au
    for graph in allundersamples:
        if graph == H:
            return True
    return False



"""This module contains clingo interaction functions"""
from __future__ import print_function
from string import Template
from gunfolds.utils.clingo import clingo
from gunfolds.utils.calc_procs import get_process_count
from gunfolds.conversions import g2clingo, rate, rasl_jclingo2g,\
     drasl_jclingo2g, clingo_preamble,\
     numbered_g2clingo, numbered_g2wclingo, encode_list_sccs

CLINGO_LIMIT = 64
PNUM = min(CLINGO_LIMIT, get_process_count(1))
CAPSIZE = 1

all_u_rasl_program = """
{edge(X,Y)} :- node(X), node(Y).
directed(X,Y,1) :- edge(X,Y).
directed(X,Y,L) :- directed(X,Z,L-1), edge(Z,Y), L <= U, u(U).
bidirected(X,Y,U) :- directed(Z,X,L), directed(Z,Y,L), node(X;Y;Z), X < Y, L < U, u(U).
countdirh(C):- C = #count { hdirected(X, Y): hdirected(X, Y), node(X), node(Y)}.
countbidirh(C):- C = #count { hbidirected(X, Y): hbidirected(X, Y), node(X), node(Y)}.
equald(L):- { directed(X,Y,L): hdirected(X,Y), node(X), node(Y) } == C, countdirh(C),u(L).
equalb(L):- { bidirected(X,Y,L): hbidirected(X,Y), node(X), node(Y) } == C, countbidirh(C),u(L).
equal(L) :- equald(L), equalb(L).
{trueu(L)} :- equal(L).
equaltest(M) :- 1 < {equal(_)}, equal(M).
min(M):- #min {MM:equaltest(MM)}=M, equaltest(_).
repeat(N):- min(M), equal(N), M<N.
:- directed(X, Y, L), not hdirected(X, Y), node(X), node(Y), trueu(L).
:- not directed(X, Y, L) , hdirected(X, Y), trueu(L).
:- bidirected(X, Y, L), not hbidirected(X, Y), node(X), node(Y), X < Y, trueu(L).
:- not bidirected(X, Y, L), hbidirected(X, Y), X < Y, trueu(L).
:- not trueu(_).
:- min(M), trueu(N), M<N.
    """

# The ASP formulation that does the heavylifting of the RASL encoding
# -------------------------------------------------------------------
# Generate powerset of all possible edges and for given edges produce directed
# edges up to the current undersampling and bidirected edges at the current
# undersampling.
drasl_program = """
    {edge1(X,Y)} :- node(X), node(Y).
    directed(X, Y, 1) :- edge1(X, Y).
    directed(X, Y, L) :- directed(X, Z, L-1), edge1(Z, Y), L <= U, u(U, _).
    bidirected(X, Y, U) :- directed(Z, X, L), directed(Z, Y, L), node(X;Y;Z), X < Y, L < U, u(U, _).
    """
# A set of no-go rules that compare the produces undersampled graph from the
# current element of the edge powerset to the input measurement timescae
# graph(s) and get rid of solutions that have a mismatch.
drasl_program += """
    :- directed(X, Y, L), not hdirected(X, Y, K), node(X;Y), u(L, K).
    :- bidirected(X, Y, L), not hbidirected(X, Y, K), node(X;Y), u(L, K), X < Y.
    :- not directed(X, Y, L), hdirected(X, Y, K), node(X;Y), u(L, K).
    :- not bidirected(X, Y, L), hbidirected(X, Y, K), node(X;Y), u(L, K), X < Y.
    """
# Filer out graphs that have already converged
# ----------------------------------------
# The next two sections are connected and are there to make sure that we do not
# count graphs more than once. Withoutthese sections if a converged or
# oscillating graph matches the measured graph H at multiple undersampling
# rates, all of them will be listed as unique solutions. The main idea is to
# first define a notequal operator, that works by checking individual edges,
# and then filter out solutions where this operator does not hold for at least
# one pair of undersampling rates.

# Turns out that the generating rules above produce all directed edges for each
# undersampling rate from 1 to u in each answer set, but only one set of
# bidirected edges that correspond to the current u in this answer. Often, it
# suffices to check repeatition of all directed edges to determine that a graph
# has converged. The main reason is persistence of bidirected edges that stay
# present once they appear at some undersampling rate. However, in some cases
# the directed edges converge before the bidirected edges do. This happens when
# the final set of directed edges generates at least one bidirected edge at the
# next step.

# The following two lines take advantage of the present history of all directed
# edges across undersampling rates up to u, and build a workaround for the
# absent history of bidirected edges. The first line is true if there was a
# fork with X and Y at the ends before the last step of undersampling. This
# would mean that the bidirected adge, if it exists at the current
# undersampling step, was already there at the last step as well. The second
# line sets notequal to true if at the current undersampling rate L a new
# bidirected edge appeared.
drasl_program += """
    pastfork(X,Y,L) :- directed(Z, X, K), directed(Z, Y, K), node(X;Y;Z), X < Y, K < L-1, uk(K), u(L, _).
    notequal(L-1,L) :- bidirected(X,Y,L), not pastfork(X,Y,L), node(X;Y), X < Y, u(L, _).
    """

# The following lines set the notequal condition for the cases when there is a
# mismatch in directed edges between rate K lower than L and L - the current
# undersampling rate.
drasl_program += """
    notequal(K,L) :- directed(X, Y, K), not directed(X, Y, L), node(X;Y), K<L, uk(K), u(L,_).
    notequal(K,L) :- not directed(X, Y, K), directed(X, Y, L), node(X;Y), K<L, uk(K), u(L,_).
    """

# This line filters out all solutions for which there is a K lower than the
# current undersampling rate and the graph at rate K equals to the one at rate
# L. This is a way to bypass our inability to specify the foroll quantifier
# directly in clingo.
drasl_program += """
    :- not notequal(K,L), K<L, uk(K), u(L,_).
    """

# The following section refuces to handle graphs if their compressed
# representation is a DAGs withuot forks. Too many options that are
# uninformative anyways and thre is no need to waste computation on them.
drasl_program += """
    nonempty(L) :- directed(X, Y, L), u(L,_).
    nonempty(L) :- bidirected(X, Y, L), u(L,_).
    :- not nonempty(L), u(L,_).
    """


def weighted_drasl_program(directed, bidirected):
    """
    Adjusts the optimization code based on the directed and bidirected priority

    :param directed: priority of directed edges in optimization
    :type directed: integer

    :param bidirected: priority of bidirected edges in optimization
        graph
    :type bidirected: integer

    :returns: optimization part of the ``clingo`` code
    :rtype: string
    """
    t = Template("""
    {edge1(X,Y)} :- node(X), node(Y).
    directed(X, Y, 1) :- edge1(X, Y).
    directed(X, Y, L) :- directed(X, Z, L-1), edge1(Z, Y), L <= U, u(U, _).
    bidirected(X, Y, U) :- directed(Z, X, L), directed(Z, Y, L), node(X;Y;Z), X < Y, L < U, u(U, _).

    :~ not directed(X, Y, L), hdirected(X, Y, W, K), node(X;Y), u(L, K). [W@$directed,X,Y]
    :~ not bidirected(X, Y, L), hbidirected(X, Y, W, K), node(X;Y), u(L, K), X < Y. [W@$bidirected,X,Y]
    :~ directed(X, Y, L), no_hdirected(X, Y, W, K), node(X;Y), u(L, K). [W@$directed,X,Y]
    :~ bidirected(X, Y, L), no_hbidirected(X, Y, W, K), node(X;Y), u(L, K), X < Y. [W@$bidirected,X,Y]

    pastfork(X,Y,L) :- directed(Z, X, K), directed(Z, Y, K), node(X;Y;Z), X < Y, K < L-1, uk(K), u(L, _).
    notequal(L-1,L) :- bidirected(X,Y,L), not pastfork(X,Y,L), node(X;Y), X < Y, u(L, _).
    notequal(K,L) :- directed(X, Y, K), not directed(X, Y, L), node(X;Y), K<L, uk(K), u(L,_).
    notequal(K,L) :- not directed(X, Y, K), directed(X, Y, L), node(X;Y), K<L, uk(K), u(L,_).
    :- not notequal(K,L), K<L, uk(K), u(L,_).
    nonempty(L) :- directed(X, Y, L), u(L,_).
    nonempty(L) :- bidirected(X, Y, L), u(L,_).
    :- not nonempty(L), u(L,_).
    """)
    return t.substitute(directed=directed, bidirected=bidirected)


def rate(u, uname='u'):
    """
    Adds under sampling rate to ``clingo`` code

    :param u: maximum under sampling rate
    :type u: integer

    :param uname: name of the parameter
    :type uname: string

    :returns: predicate for under sampling rate
    :rtype: string
    """
    s = "1 {" + uname + "(1.."+str(u)+")} 1."
    return s


def drate(u, gnum, weighted=False):
    """
    Replaces ``rate`` if there are multiple under sampled inputs

    :param u: maximum under sampling rate
    :type u: integer

    :param gnum: number of under sampled inputs
    :type gnum: integer

    :param weighted: whether the input graphs are weighted or
        precize.  If `True` but no weight matrices are provided -
        all weights are set to `1`
    :type weighted: boolean

    :returns: ``clingo`` code for under sampling with multiple under sampled inputs
    :rtype: string
    """
    s = f"1 {{u({int(weighted)+1}..{u}, {gnum})}} 1."
    return s


def rasl_command(g, urate=0):
    """
    Given a graph generates ``clingo`` code

    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)

    :param urate: maximum undersampling rate to consider
    :type urate: integer

    :returns: ``clingo`` code
    :rtype: string
    """
    if not urate:
        urate = 1+3*len(g)
    command = g2clingo(g) + ' ' + rate(urate) + ' '
    command += '{edge(X,Y)} :- node(X), node(Y). ' + all_u_rasl_program + ' '
    command += "#show edge/2. "
    command += "#show trueu/1. "
    command += "#show min/1."
    command = command.encode().replace(b"\n", b" ")
    return command


def glist2str(g_list, weighted=False, dm=None, bdm=None):
    """
    Converts list of graphs into ``clingo`` predicates

    :param g_list: a list of graphs that are undersampled versions of
        the same system
    :type g_list: list of dictionaries (``gunfolds`` graphs)

    :param weighted: whether the input graphs are weighted or
        precize.  If `True` but no weight matrices are provided -
        all weights are set to `1`
    :type weighted: boolean

    :param dm: a list of n-by-n 2-d square matrix of the weights for
        directed edges of each input n-node graph
    :type dm: list of numpy arrays

    :param bdm: a list of *symmetric* n-by-n 2-d square matrix of the
        weights for bidirected edges of each input n-node graph
    :type bdm: list of numpy arrays

    :returns: ``clingo`` predicates as a string
    :rtype: string
    """
    if dm is None:
        dm = [None]*len(g_list)
    else:
        dm = [nd.astype('int') for nd in dm]
    if bdm is None:
        bdm = [None]*len(g_list)
    else:
        bdm = [nd.astype('int') for nd in bdm]
    s = ''
    for count, (g, D, B) in enumerate(zip(g_list, dm, bdm)):
        if weighted:
            s += numbered_g2wclingo(g, count+1, directed_weights_matrix=D, bidirected_weights_matrix=B) + ' '
        else:
            s += numbered_g2clingo(g, count+1) + ' '
    return s


def drasl_command(g_list, max_urate=0, weighted=False, scc=False, scc_members=None, dm=None, bdm=None, edge_weights=(1, 1)):
    """
    Given a list of graphs generates ``clingo`` codes

    :param g_list: a list of graphs that are undersampled versions of
        the same system
    :type g_list: list of dictionaries (``gunfolds`` graphs)

    :param max_urate: maximum under sampling rate
    :type max_urate: integer

    :param weighted: whether the input graphs are weighted or
        precize.  If ``True`` but no weight matrices are provided -
        all weights are set to ``1``
    :type weighted: boolean

    :param scc: whether to assume that each SCC in the input graph is
        either a singleton or have ``gcd=1``.  If `True` a much more
        efficient algorithm is employed.
    :type scc: (GUESS)boolean

    :param scc_members: a list of sets for nodes in each SCC
    :type scc_members: list

    :param dm: a list of n-by-n 2-d square matrix of the weights for
        directed edges of each input n-node graph
    :type dm: list of numpy arrays

    :param bdm: a list of *symmetric* n-by-n 2-d square matrix of the
        weights for bidirected edges of each input n-node graph
    :type bdm: list of numpy arrays

    :param edge_weights: a tuple of 2 values, the first is importance of matching
        directed weights when solving optimization problem and the second is for bidirected.
    :type edge_weights: tuple with 2 elements

    :returns: clingo code as a string
    :rtype: string
    """
    if dm is not None:
        dm = [nd.astype('int') for nd in dm]
    if bdm is not None:
        bdm = [nd.astype('int') for nd in bdm]

    assert len({len(g) for g in g_list}) == 1, "Input graphs have variable number of nodes!"

    if not max_urate:
        max_urate = 1+3*len(g_list[0])
    n = len(g_list)
    command = clingo_preamble(g_list[0])
    if scc:
        command += encode_list_sccs(g_list, scc_members)
    command += f"dagl({len(g_list[0])-1}). "
    command += glist2str(g_list, weighted=weighted, dm=dm, bdm=bdm) + ' '   # generate all graphs
    command += 'uk(1..'+str(max_urate)+').' + ' '
    command += ' '.join([drate(max_urate, i+1, weighted=weighted) for i in range(n)]) + ' '
    command += weighted_drasl_program(edge_weights[0], edge_weights[1]) if weighted else drasl_program
    command += f":- M = N, {{u(M, 1..{n}); u(N, 1..{n})}} == 2, u(M, _), u(N, _). "
    command += "#show edge1/2. "
    command += "#show u/2."
    command = command.encode().replace(b"\n", b" ")
    return command


def drasl(glist, capsize=CAPSIZE, timeout=0, urate=0, weighted=False, scc=False, scc_members=None, dm=None,
          bdm=None, pnum=PNUM, edge_weights=(1, 1), configuration="crafty", optim='optN'):
    """
    Compute all candidate causal time-scale graphs that could have
    generated all undersampled graphs at all possible undersampling
    rates up to ``urate`` in ``glist`` each at an unknown undersampling
    rate.

    :param glist: a list of graphs that are undersampled versions of
        the same system
    :type glist: list of dictionaries (``gunfolds`` graphs)

    :param capsize: maximum number of candidates to return
    :type capsize: integer

    :param timeout: timeout in seconds after which to interrupt
        computation (0 - no limit)
    :type timeout: integer

    :param urate: maximum undersampling rate to consider
    :type urate: integer

    :param weighted: whether the input graphs are weighted or
        imprecize.  If ``True`` but no weight matrices are provided -
        all weights are set to ``1``
    :type weighted: boolean

    :param scc: whether to assume that each SCC in the input graph is
        either a singleton or have ``gcd=1``.  If ``True`` a much more
        efficient algorithm is employed.
    :type scc: boolean

    :param scc_members: a list of sets for nodes in each SCC
    :type scc_members: list

    :param dm: a list of n-by-n 2-d square matrix of the weights for
        directed edges of each input n-node graph
    :type dm: list of numpy arrays

    :param bdm: a list of *symmetric* n-by-n 2-d square matrix of the
        weights for bidirected edges of each input n-node graph
    :type bdm: list of numpy arrays

    :param pnum: number of parallel threads to run ``clingo`` on
    :type pnum: integer

    :param edge_weights: a tuple of 2 values, the first is importance
        of matching directed weights when solving optimization problem and the second is for bidirected.
    :type edge_weights: tuple with 2 elements

    :param configuration: Select configuration based on problem type

        - ``frumpy`` : Use conservative defaults
        - ``jumpy`` : Use aggressive defaults
        - ``tweety`` : Use defaults geared towards asp problems
        - ``handy`` : Use defaults geared towards large problems
        - ``crafty`` : Use defaults geared towards crafted problems
        - ``trendy`` : Use defaults geared towards industrial problems
    :type configuration: string

    :param optim: a comma separated string containing configuration for optimization algorithm and optionally a bound [<arg>[, <bound>]]
        
        - <arg> : <mode {opt|enum|optN|ignore}>
            - ``opt`` : Find optimal model
            - ``enum`` : Find models with costs <= <bound>
            - ``optN`` : Find optimum, then enumerate optimal models
            - ``ignore`` : Ignore optimize statements
        - <bound> : Set initial bound for objective function(s)
    :type optim: string
    
    :returns: results of parsed equivalent class
    :rtype: dictionary
    """
    if dm is not None:
        dm = [nd.astype('int') for nd in dm]
    if bdm is not None:
        bdm = [nd.astype('int') for nd in bdm]
    if not isinstance(glist, list):
        glist = [glist]
    return clingo(drasl_command(glist, max_urate=urate, weighted=weighted,
                                scc=scc, scc_members=scc_members, dm=dm, bdm=bdm, edge_weights=edge_weights),
                  capsize=capsize, convert=drasl_jclingo2g, configuration=configuration,
                  timeout=timeout, exact=not weighted, pnum=pnum, optim=optim)


def rasl(g, capsize, timeout=0, urate=0, pnum=None, configuration="tweety"):
    """
    :param g: ``gunfolds`` graph
    :type g: dictionary (``gunfolds`` graphs)

    :param capsize: maximum number of candidates to return
    :type capsize: integer

    :param timeout: timeout in seconds after which to interrupt
        computation (0 - no limit)
    :type timeout: integer

    :param urate: maximum undersampling rate to consider
    :type urate: integer

    :param pnum: number of parallel threads to run ``clingo`` on
    :type pnum: integer

    :param configuration: Select configuration based on problem type

        - ``frumpy`` : Use conservative defaults
        - ``jumpy`` : Use aggressive defaults
        - ``tweety`` : Use defaults geared towards asp problems
        - ``handy`` : Use defaults geared towards large problems
        - ``crafty`` : Use defaults geared towards crafted problems
        - ``trendy`` : Use defaults geared towards industrial problems
    :type configuration: string

    :returns: results of parsed equivalent class
    :rtype: dictionary
    """
    return clingo(rasl_command(g, urate=urate), capsize=capsize, configuration=configuration, convert=rasl_jclingo2g, timeout=timeout, pnum=pnum)



""" This module contains clingo interaction functions """
from __future__ import print_function
from gunfolds.conversions import drasl_jclingo2g
import clingo as clngo
from gunfolds.utils.calc_procs import get_process_count

CLINGO_LIMIT = 64
PNUM = min(CLINGO_LIMIT, get_process_count(1))
CAPSIZE = 1

def run_clingo(command,
               exact=True,
               timeout=0,
               capsize=CAPSIZE,
               configuration="tweety",
               pnum=PNUM,
               optim='optN'):
    """
    Open sub-process and run clingo

    :param command: Completed clingo code
    :type command: string

    :param exact: If true, run clingo in exact mode. If false, run clingo in optimization mode
    :type exact: boolean

    :param timeout: timeout in seconds after which to interrupt
        computation (0 - no limit)
    :type timeout: integer

    :param capsize: maximum number of candidates to return
    :type capsize: integer

    :param configuration: Select configuration based on problem type

        - ``frumpy`` : Use conservative defaults
        - ``jumpy`` : Use aggressive defaults
        - ``tweety`` : Use defaults geared towards asp problems
        - ``handy`` : Use defaults geared towards large problems
        - ``crafty`` : Use defaults geared towards crafted problems
        - ``trendy`` : Use defaults geared towards industrial problems
    :type configuration: string


    :param pnum: number of parallel threads to run clingo on
    :type pnum: integer

    :param optim: a comma separated string containing configuration for optimization algorithm and optionally a bound [<arg>[, <bound>]]
        
        - <arg> : <mode {opt|enum|optN|ignore}>
            - ``opt`` : Find optimal model
            - ``enum`` : Find models with costs <= <bound>
            - ``optN`` : Find optimum, then enumerate optimal models
            - ``ignore`` : Ignore optimize statements
        - <bound> : Set initial bound for objective function(s)
    :type optim: string

    :returns: results of equivalent class
    :rtype: dictionary
    """
    assert len(optim.split(',')) < 3, "optim option only takes 1 or 2 comma-separated parameters"

    clingo_control = ["--warn=no-atom-undefined", "--configuration=", configuration, "-t", str(int(pnum)) + ",split", "-n", str(capsize)]
    ctrl = clngo.Control(clingo_control)
    if not exact:
        ctrl.configuration.solve.opt_mode = optim
    ctrl.add("base", [], command.decode())
    ctrl.ground([("base", [])])
    models = []
    with ctrl.solve(yield_=True, async_=True) as handle:
        for model in handle:
            models.append(([str(atom) for atom in model.symbols(shown=True)], model.cost))
    cost = ctrl.statistics["summary"]["costs"]
    num_opt = ctrl.statistics["summary"]["models"]["optimal"]
    if not exact:
        if num_opt == 0.0:
            return {}, cost
        else:
            return models, cost 
    return models, cost
    

def clingo(command, exact=True,
           convert=drasl_jclingo2g,
           timeout=0,
           capsize=CAPSIZE,
           configuration="crafty",
           optim='optN',
           pnum=PNUM):
    """
    Runs ``run_clingo`` and returns parsed equivalent class

    :param command: Completed clingo code
    :type command: string

    :param exact: If true, run clingo in exact mode. If false, run clingo in optimization mode
    :type exact: boolean

    :param convert: result parsing protocol
    :type convert: function

    :param timeout: timeout in seconds after which to interrupt
        computation (0 - no limit)
    :type timeout: integer

    :param capsize: maximum number of candidates to return
    :type capsize: integer

    :param configuration: Select configuration based on problem type

        - ``frumpy`` : Use conservative defaults
        - ``jumpy`` : Use aggressive defaults
        - ``tweety`` : Use defaults geared towards asp problems
        - ``handy`` : Use defaults geared towards large problems
        - ``crafty`` : Use defaults geared towards crafted problems
        - ``trendy`` : Use defaults geared towards industrial problems
    :type configuration: string

    :param optim: a comma separated string containing configuration for optimization algorithm and optionally a bound [<arg>[, <bound>]]
        
        - <arg> : <mode {opt|enum|optN|ignore}>
            - ``opt`` : Find optimal model
            - ``enum`` : Find models with costs <= <bound>
            - ``optN`` : Find optimum, then enumerate optimal models
            - ``ignore`` : Ignore optimize statements
        - <bound> : Set initial bound for objective function(s)
    :type optim: string

    :param pnum: number of parallel threads to run clingo on
    :type pnum: integer

    :returns: results of parsed equivalent class
    :rtype: dictionary
    """
    result = run_clingo(command,
                        exact=exact,
                        timeout=timeout,
                        capsize=capsize,
                        configuration=configuration,
                        pnum=pnum,
                        optim=optim)
    if result[0] == {} or result[0] == []:
        return {}
    else:
        if not exact:
            r = {(convert(value[0]), value[1][0]) for value in result[0]}
        else:
            r = {convert(value[0]) for value in result[0]}
    return r



""" BFS implementation of subgraph and supergraph Gu to G1 algorithm """
import gunfolds.utils.bfutils as bfu
from gunfolds.conversions import g2num, ug2num, num2CG
import gunfolds.utils.graphkit as gk
import gunfolds.utils.simpleloops as sls
from gunfolds.utils import load_data
from itertools import combinations
import numpy as np
import operator
from progressbar import ProgressBar, Bar
from functools import reduce


def prune_conflicts(H, g, elist):
    """Checks if adding an edge from the list to graph g causes a
    conflict with respect to H and if it does removes the edge
    from the list

    :param H: the undersampled graph
    :type H:
    
    :param g: a graph under construction
    :type g:
        
    :param elist: list of edges to check
    :type elist:
    
    :returns: 
    :rtype:
    """
    l = []
    for e in elist:
        gk.addanedge(g, e)
        if not bfu.call_u_conflicts(g, H):
            l.append(e)
        gk.delanedge(g, e)
    return l


def e2num(e, n):
    """
    :param e:
    :type e:
    
    :param n:
    :type n:
    
    :returns: 
    :rtype:
    """
    return 1 << (n*n + n - e[0]*n - e[1])


def le2num(elist, n):
    """
    :param elist: list of edges to check
    :type elist:
    
    :param n:
    :type n:
    
    :returns: 
    :rtype:
    """
    num = 0
    for e in elist:
        num |= e2num(e, n)
    return num


def ekey2e(ekey, n):
    """
    :param ekey:
    :type ekey:
    
    :param n:
    :type n:
    
    :returns: 
    :rtype:
    """
    idx = np.unravel_index(n * n - ekey .bit_length() - 1 + 1, (n, n))
    idx = tuple([x + 1 for x in idx])
    return ('%i %i' % idx).split(' ')


def cacheconflicts(num, cache):
    """Given a number representation of a graph and an iterable of
    conflicting subgraphs return True if the graph conflicts with any
    of them and false otherwise

    :param num: the number representation of a graph
    :type num:
    
    :param cache: an iterable of number representations of conflicting graphs
    :type cache:
          
    :returns: 
    :rtype:
    """
    for c in cache:
        if num & c == c:
            return True
    return False


class nobar:

    def update(self):
        return None

    def finish(self):
        return None


def start_progress_bar(iter, n, verbose=True):
    """
    :param iter:
    :type iter:
    
    :param n:
    :type n:
    
    :param verbose:
    :type verbose:
    
    :returns: 
    :rtype:
    """
    if verbose:
        pbar = ProgressBar(widgets=['%3s' % str(iter) +
                           '%10s' % str(n) + ' ', Bar('-'), ' '],
                           maxval=n).start()
    else:
        pbar = nobar()
    return pbar


def add2set_loop(ds, H, cp, ccf, iter=1, verbose=True,
                 capsize=100, currsize=0):
    """
    :param ds:
    :type ds:
    
    :param H:
    :type H:
    
    :param cp:
    :type cp:
    
    :param ccf:
    :type ccf:
    
    :param iter:
    :type iter:
    
    :param verbose:
    :type verbose:
    
    :param capsize:
    :type capsize:
    
    :param currsize:
    :type currsize:
    
    :returns: 
    :rtype:
    """
    n = len(H)
    dsr = {}
    s = set()
    ss = set()
    pbar = start_progress_bar(iter, len(ds), verbose=verbose)
    c = 0

    for gnum in ds:
        c += 1
        pbar.update(c)
        gset = set()
        eset = set()
        for sloop in ds[gnum]:
            if sloop & gnum == sloop:
                continue
            num = sloop | gnum
            if sloop in ccf and skip_conflictors(num, ccf[sloop]):
                continue
            if not (num in s):
                g = num2CG(num, n)
                if not bfu.call_u_conflicts(g, H):
                    s.add(num)
                    gset.add((num, sloop))
                    eset.add(sloop)
                    if bfu.call_u_equals(g, H):
                        ss.add(num)
                        if capsize <= len(ss) + currsize:
                            return dsr, ss

        for gn, e in gset:
            if e in cp:
                dsr[gn] = eset - cp[e] - set([e])
            else:
                dsr[gn] = eset - set([e])

    pbar.finish()
    return dsr, ss


def add2set_(ds, H, cp, ccf, iter=1, verbose=True, capsize=100):
    """
    :param ds:
    :type ds:
    
    :param H:
    :type H:
    
    :param cp:
    :type cp:
    
    :param ccf:
    :type ccf:
    
    :param iter:
    :type iter:
    
    :param verbose:
    :type verbose:
    
    :param capsize:
    :type capsize:
    
    :returns: 
    :rtype:
    """
    n = len(H)
    n2 = n * n + n
    dsr = {}
    s = set()
    ss = set()

    pbar = start_progress_bar(iter, len(ds), verbose=verbose)

    c = 0

    for gnum in ds:
        g = num2CG(gnum, n)
        c += 1
        pbar.update(c)
        glist = []
        elist = []
        eset = set()
        for e in ds[gnum]:
            if not e[1] in g[e[0]]:
                gk.addanedge(g, e)
                num = g2num(g)
                ekey = (1 << (n2 - e[0]*n - e[1]))
                if ekey in ccf and skip_conflictors(num, ccf[ekey]):
                    gk.delanedge(g, e)
                    continue
                if not (num in s):
                    s.add(num)
                    if not bfu.call_u_conflicts(g, H):
                        # cf, gl2 = bfu.call_u_conflicts2(g, H)
                        # if not cf:
                        glist.append((num, ekey))
                        elist.append(e)
                        eset.add(ekey)
                        if bfu.call_u_equals(g, H):
                            ss.add(num)
                            # if bfu.call_u_equals2(g, gl2, H): ss.add(num)
                        if capsize <= len(ss):
                            break
                gk.delanedge(g, e)

        for gn, e in glist:
            if e in cp:
                dsr[gn] = [ekey2e(k, n) for k in eset - cp[e]]
            else:
                dsr[gn] = elist
        if capsize <= len(ss):
            return dsr, ss

    pbar.finish()
    return dsr, ss


def skip_conflictors(gnum, ccf):
    """
    :param gnum:
    :type gnum:
    
    :param ccf:
    :type ccf:
    
    :returns: 
    :rtype:
    """
    pss = False
    for xx in ccf:
        if xx & gnum == xx:
            pss = True
            break
    return pss


def bconflictor(e, H):
    """
    :param e:
    :type e:
    
    :param H:
    :type H:
    
    :returns: 
    :rtype:
    """
    n = len(H)
    s = set()
    for v in H:
        s.add(e2num((v, e[0]), n) | e2num((v, e[1]), n))
    return s


def conflictor(e, H):
    """
    :param e:
    :type e:
    
    :param H:
    :type H:
    
    :returns: 
    :rtype:
    """
    n = len(H)

    def pairs(e):
        ekey = e2num(e, n)
        return [ekey | e2num((e[0], e[0]), n),
                ekey | e2num((e[1], e[1]), n)]

    def trios(e, H):
        s = set()
        for v in H:
            if not (v in e):
                s.add(e2num((e[0], v), n) |
                      e2num((v, e[1]), n) |
                      e2num((e[1], e[1]), n))
                s.add(e2num((e[0], e[0]), n) |
                      e2num((e[0], v), n) |
                      e2num((v, e[1]), n))
                s.add(e2num((e[0], v), n) |
                      e2num((v, v), n) |
                      e2num((v, e[1]), n))
        return s

    return trios(e, H).union(pairs(e))


def conflictor_set(H):
    """
    :param H:
    :type H:
    
    :returns: 
    :rtype:
    """
    s = set()
    for x in gk.inedgelist(H):
        s = s | conflictor(x, H)
    for x in gk.inbedgelist(H):
        s = s | bconflictor(x, H)
    return s


def conflictors(H):
    """
    :param H:
    :type H:
    
    :returns: 
    :rtype:
    """
    s = conflictor_set(H)
    ds = {}
    num = reduce(operator.or_, s)
    for i in range(len(bin(num))-2):
        if num & 1 << i:
            ds[1 << i] = [x for x in s if x & (1 << i)]
    return ds


def prune_loops(loops, H):
    """
    :param loops:
    :type loops:
    
    :param H:
    :type H:
    
    :returns: 
    :rtype:
    """
    l = []
    n = len(H)
    for loop in loops:
        g = num2CG(loop, n)
        if not bfu.call_u_conflicts_d(g, H):
            l.append(loop)
    return l


def lconflictors(H, sloops=None):
    """
    :param H:
    :type H:
    
    :param sloops:
    :type sloops:
    
    :returns: 
    :rtype:
    """
    if not sloops:
        sloops = prune_loops(allsloops(len(H)), H)
    s = conflictor_set(H)
    ds = {}
    num = reduce(operator.or_, s)
    for i in range(len(bin(num))-2):
        if num & 1 << i:
            cset = [x for x in s if x & (1 << i)]
            for sloop in sloops:
                if sloop & 1 << i:
                    ds.setdefault(sloop, []).extend(cset)
    return ds


def confpairs(H):
    """
    :param H:
    :type H:
    
    :returns: 
    :rtype:
    """
    n = len(H)
    g = {n: {} for n in H}
    d = {}

    edges = gk.edgelist(gk.complement(g))
    edges = prune_conflicts(H, g, edges)

    for p in combinations(edges, 2):
        gk.addedges(g, p)
        if bfu.call_u_conflicts(g, H):
            n1 = e2num(p[0], n)
            n2 = e2num(p[1], n)
            d.setdefault(n1, set()).add(n2)
            d.setdefault(n2, set()).add(n1)
        gk.deledges(g, p)

    return d


def lconfpairs(H, cap=10, sloops=None):
    """
    :param H:
    :type H:
    
    :param cap:
    :type cap:
    
    :param sloops:
    :type sloops:
    
    :returns: 
    :rtype:
    """
    n = len(H)
    d = {}
    if not sloops:
        sloops = prune_loops(allsloops(len(H)), H)
    c = 0
    for p in combinations(sloops, 2):
        g = num2CG(p[0] | p[1], n)
        if bfu.call_u_conflicts(g, H):
            d.setdefault(p[0], set()).add(p[1])
            d.setdefault(p[1], set()).add(p[0])
        if c >= cap:
            break
        c += 1
    return d


def iteqclass(H, verbose=True, capsize=100):
    """
    Find all graphs in the same equivalence class with respect to
    graph H and any undesampling rate.
    
    :param H:
    :type H:

    :param verbose:
    :type verbose:
    
    :param capsize:
    :type capsize:
    
    :returns: 
    :rtype:
    """
    if bfu.is_sclique(H):
        print('not running on superclique')
        return None
    g = {n: {} for n in H}
    s = set()
    Hnum = ug2num(H)
    if Hnum[1] == 0:
        s.add(Hnum[0])

    cp = confpairs(H)
    ccf = conflictors(H)

    edges = gk.edgelist(gk.complement(g))
    ds = {g2num(g): edges}

    if verbose:
        print('%3s' % 'i'+'%10s' % ' graphs')
    for i in range(len(H) ** 2):
        ds, ss = add2set_(ds, H, cp, ccf, iter=i,
                          verbose=verbose,
                          capsize=capsize)
        s = s | ss
        if capsize <= len(ss):
            break
        if not ds:
            break

    return s


def liteqclass(H, verbose=True, capsize=100, asl=None):
    """
    Find all graphs in the same equivalence class with respect to
    graph H and any undesampling rate.
    
    :param H:
    :type H:

    :param verbose:
    :type verbose:
    
    :param capsize:
    :type capsize:
    
    :param asl:
    :type asl:
    
    :returns: 
    :rtype:
    """
    if bfu.is_sclique(H):
        print('not running on superclique')
        return set([-1])
    s = set()
    cp = lconfpairs(H)
    if asl:
        sloops = asl
    else:
        sloops = prune_loops(allsloops(len(H)), H)
    ccf = lconflictors(H, sloops=sloops)
    ds = {0: sloops}
    if verbose:
        print('%3s' % 'i'+'%10s' % ' graphs')
    i = 0
    while ds:
        ds, ss = add2set_loop(ds, H, cp, ccf, iter=i,
                              verbose=verbose,
                              capsize=capsize,
                              currsize=len(s))
        s = s | ss
        i += 1
        if capsize <= len(s):
            break
    return s


def skip_conflict(g1, g2, ds):
    """
    :param g1:
    :type g1:

    :param g2:
    :type g2:
    
    :param ds:
    :type ds:
    
    :returns: 
    :rtype:
    """
    pss = False
    for ekey in ds:
        if (g1 & ekey) == ekey:
            if ekey in ds and cacheconflicts(g2, ds[ekey]):
                pss = True
                break
    return pss


def loop2graph(l, n):
    """
    :param l:
    :type l: (GUESS)integer

    :param n:
    :type n: (GUESS)integer
    
    :returns: 
    :rtype:
    """
    g = {i: {} for i in range(1, n+1)}
    for i in range(len(l) - 1):
        g[l[i]][l[i + 1]] = 1
    g[l[-1]][l[0]] = 1
    return g


def perm_circular(l, cp=None):
    """
    :param l:
    :type l: (GUESS)integer

    :param cp:
    :type cp:
    
    :returns: 
    :rtype:
    """
    if cp is None:
        # Delay import until its used
        cp = load_data.circp
    r = []
    n = len(l)
    for e in cp[n]:
        r.append([l[i] for i in e])
    return r


def gen_loops(n):
    """
    :param n:
    :type n: (GUESS)integer
    
    :returns: 
    :rtype:
    """
    l = [i for i in range(1, n + 1)]
    s = []
    for i in range(1, n + 1):
        for e in combinations(l, i):
            s.extend(perm_circular(e))
    return s


def allsloops(n, asl=None):
    """
    :param n:
    :type n: (GUESS)integer
    
    :param asl:
    :type asl:
    
    :returns: 
    :rtype:
    """
    if asl is None:
        asl = load_data.alloops
    if asl:
        return asl[n]
    s = []
    l = gen_loops(n)
    for e in l:
        s.append(g2num(loop2graph(e, n)))
    return s


def reverse(H, verbose=True, capsize=1000):
    """
    :param H:
    :type H:

    :param verbose:
    :type verbose:
    
    :param capsize:
    :type capsize:
    
    :returns: 
    :rtype:
    """
    n = len(H)
    s = set()

    g = gk.superclique(n)
    sloops = set(allsloops(n))

    ds = {g2num(g): sloops}

    if verbose:
        print('%3s' % 'i'+'%10s' % ' graphs')
    i = 0
    while ds:
        ds, ss = del_loop(ds, H, iter=i,
                          verbose=verbose,
                          capsize=capsize)
        s = s | ss
        i += 1
        if capsize <= len(s):
            break

    return s

# ----------------------


def build_loop_step(ds, loop, n, iter=1):
    """
    :param ds:
    :type ds:

    :param loop:
    :type loop:
    
    :param n:
    :type n:
    
    :param iter:
    :type iter:
    
    :returns: 
    :rtype:
    """
    dsr = {}
    s = set()
    ss = set()

    pbar = start_progress_bar(iter, len(ds))

    c = 0

    for gnum in ds:
        c += 1
        pbar.update(c)
        gset = set()
        eset = set()
        for sloop in ds[gnum]:
            num = sloop | gnum
            if not (num in s):
                g = num2CG(num, n)
                s.add(num)
                if bfu.forms_loop(g, loop):
                    ss.add(num)
                else:
                    gset.add((num, sloop))
                    eset.add(sloop)

        for gn, e in gset:
            dsr[gn] = eset - set([e])

    pbar.finish()
    return dsr, ss


def forward_loop_match(loop, n):
    """
    Start with an empty graph and keep adding simple loops until
    the loop is generated at some undersampling rate

    :param loop: binary encoding of the loop
    :type loop:
    
    :param n: number of nodes in the graph
    :type n:
    
    :returns: 
    :rtype:
    """
    s = set()
    sloops = allsloops(n)
    ds = {0: sloops}

    i = 0
    while ds:
        ds, ss = build_loop_step(ds, loop, n, iter=i)
        s = s | ss
        i += 1

    return s


def delAloop(g, loop):
    """
    :param g: (GUESS)graph that generates the loop
    :param g:
    
    :param loop: (GUESS)the reference loop
    :type loop:
    
    :returns: 
    :rtype:
    """
    n = len(g)
    l = []
    l = [g2num(loop2graph(s, n)) for s in sls.simple_loops(g, 0)]
    l = [num for num in l if not num == loop]
    print(loop, ': ',  l)
    return num2CG(reduce(operator.or_, l), n)


def reverse_loop_match(g, loop):
    """
    Start with a graph and keep removing loops while the loop is still matched

    :param g: graph that generates the loop
    :param g:
    
    :param loop: the reference loop
    :type loop:
    
    :returns: 
    :rtype:
    """
    s = set()
    n = len(g)

    def prune(g):
        cannotprune = True
        for l in sls.simple_loops(gk.digonly(g), 0):
            gg = delAloop(g, g2num(loop2graph(l, n)))
            if bfu.forms_loop(gg, loop):
                cannotprune = False
                prune(gg)
        if cannotprune:
            print('one')
            s.add(g)

    prune(g)
    return s


def reverse_edge_match(g, loop):
    """
    Start with a graph and keep removing loops while the loop is still matched

    :param g: graph that generates the loop
    :param g:
    
    :param loop: the reference loop
    :type loop:
    
    :returns: 
    :rtype:
    """
    s = set()
    n = len(g)

    def prune(g):
        cannotprune = True
        for l in gk.edgelist(gk.digonly(g)):
            gk.delanedge(g, l)
            if bfu.forms_loop(g, loop):
                cannotprune = False
                prune(g)
            gk.addanedge(g, l)
        if cannotprune:
            s.add(g2num(g))

    prune(g)
    return s


def matchAloop(loop, n):
    """Returns a set of minimal graphs that generate this loop

    :param loop: binary encoding of the loop
    :type loop:
    
    :param n: number of nodes in the graph
    :type n:
    
    :returns: 
    :rtype:
    """
    s = set()
    l = forward_loop_match(loop, n)
    print(len(l))
    for g in l:
        s = s | reverse_edge_match(num2CG(g, n), loop)

    return s

# ----------------------


def del_loop(ds, H, iter=0):
    """
    :param ds:
    :type ds:
    
    :param H:
    :type H:
    
    :param iter:
    :type iter:
    
    :returns: 
    :rtype:
    """
    n = len(H)

    dsr = {}
    ss = set()
    print(iter,)
    for gnum in ds:
        gset = []
        s = set()
        for sloop in ds[gnum]:
            rset = ds[gnum] - set([sloop])
            num = reduce(operator.or_, rset)
            if not (num in s):
                g = num2CG(num, n)
                if bfu.overshoot(g, H):
                    s.add(num)
                    gset.append((num, rset))

        if gset == []:
            print('.',)
            ss.add(gnum)

        for gn in gset:
            dsr[gn[0]] = gn[1]
    print('')
    return dsr, ss


def main():
    g = gk.ringmore(6, 1)
    H = bfu.undersample(g, 1)
    ss = liteqclass(H)
    print(ss)


if __name__ == "__main__":
    main()
